# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import galsim
from galsim_test_helpers import *

@timer
def test_galsim_error():
    """Test basic usage of GalSimError
    """
    err = galsim.GalSimError("Test")
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test"
    assert isinstance(err, RuntimeError)
    check_pickle(err)


@timer
def test_galsim_value_error():
    """Test basic usage of GalSimValueError
    """
    value = 2.3
    err = galsim.GalSimValueError("Test", value)
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test Value 2.3"
    assert err.value == value
    assert err.allowed_values == None
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, ValueError)
    check_pickle(err)

    err = galsim.GalSimValueError("Test", value, (0,1,2))
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test Value 2.3 not in (0, 1, 2)"
    assert err.value == value
    assert err.allowed_values == (0,1,2)
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, ValueError)
    check_pickle(err)


@timer
def test_galsim_key_error():
    """Test basic usage of GalSimKeyError
    """
    key = 'foo'
    err = galsim.GalSimKeyError("Test", key)
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test Key foo"
    assert err.key == key
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, KeyError)
    check_pickle(err)


@timer
def test_galsim_index_error():
    """Test basic usage of GalSimIndexError
    """
    index = 3
    err = galsim.GalSimIndexError("Test", index)
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test Index 3"
    assert err.index == index
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, IndexError)
    check_pickle(err)


@timer
def test_galsim_range_error():
    """Test basic usage of GalSimRangeError
    """
    value = 2.3
    err = galsim.GalSimRangeError("Test", value, 0, 1)
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test Value 2.3 not in range [0, 1]"
    assert err.value == value
    assert err.min == 0
    assert err.max == 1
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, ValueError)
    check_pickle(err)

    err = galsim.GalSimRangeError("Test", value, 10)
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test Value 2.3 not in range [10, None]"
    assert err.value == value
    assert err.min == 10
    assert err.max == None
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, ValueError)
    check_pickle(err)


@timer
def test_galsim_bounds_error():
    """Test basic usage of GalSimBoundsError
    """
    pos = galsim.PositionI(0,0)
    bounds = galsim.BoundsI(1,10,1,10)
    err = galsim.GalSimBoundsError("Test", pos, bounds)
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test galsim.PositionI(0,0) not in galsim.BoundsI(1,10,1,10)"
    assert err.pos == pos
    assert err.bounds == bounds
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, ValueError)
    check_pickle(err)


@timer
def test_galsim_undefined_bounds_error():
    """Test basic usage of GalSimUndefinedBoundsError
    """
    err = galsim.GalSimUndefinedBoundsError("Test")
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test"
    assert isinstance(err, galsim.GalSimError)
    check_pickle(err)


@timer
def test_galsim_immutable_error():
    """Test basic usage of GalSimImmutableError
    """
    im = galsim.ImageD(np.array([[0]]), make_const=True)
    err = galsim.GalSimImmutableError("Test", im)
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test Image: galsim.Image(bounds=galsim.BoundsI(1,1,1,1), wcs=None, dtype=numpy.float64)"
    assert err.image == im
    assert isinstance(err, galsim.GalSimError)
    check_pickle(err)


@timer
def test_galsim_incompatible_values_error():
    """Test basic usage of GalSimIncompatibleValuesError
    """
    err = galsim.GalSimIncompatibleValuesError("Test", a=1, b=2)
    print('str = ',str(err))
    print('repr = ',repr(err))
    # This isn't completely deterministic across python versions.
    str_possibilities = ["Test Values {'a': 1, 'b': 2}",
                         "Test Values {'b': 2, 'a': 1}"]
    assert str(err) in str_possibilities
    assert err.values == dict(a=1, b=2)
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, ValueError)
    assert isinstance(err, TypeError)
    check_pickle(err)


@timer
def test_galsim_sed_error():
    """Test basic usage of GalSimSEDError
    """
    sed = galsim.SED('1', wave_type='nm', flux_type='fphotons')
    err = galsim.GalSimSEDError("Test", sed)
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test SED: galsim.SED('1', redshift=0.0)"
    assert err.sed == sed
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, TypeError)
    check_pickle(err)


@timer
def test_galsim_hsm_error():
    """Test basic usage of GalSimHSMError
    """
    err = galsim.GalSimHSMError("Test")
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test"
    assert isinstance(err, galsim.GalSimError)
    check_pickle(err)



@timer
def test_galsim_config_error():
    """Test basic usage of GalSimConfigError
    """
    err = galsim.GalSimConfigError("Test")
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test"
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, ValueError)
    check_pickle(err)


@timer
def test_galsim_config_value_error():
    """Test basic usage of GalSimConfigValueError
    """
    value = 2.3
    err = galsim.GalSimConfigValueError("Test", value)
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test Value 2.3"
    assert err.value == value
    assert err.allowed_values == None
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, galsim.GalSimConfigError)
    assert isinstance(err, ValueError)
    check_pickle(err)

    err = galsim.GalSimConfigValueError("Test", value, (0,1,2))
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test Value 2.3 not in (0, 1, 2)"
    assert err.value == value
    assert err.allowed_values == (0,1,2)
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, galsim.GalSimConfigError)
    assert isinstance(err, ValueError)
    check_pickle(err)


@timer
def test_galsim_notimplemented_error():
    """Test basic usage of GalSimNotImplementedError
    """
    err = galsim.GalSimNotImplementedError("Test")
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test"
    assert isinstance(err, galsim.GalSimError)
    assert isinstance(err, NotImplementedError)
    check_pickle(err)


@timer
def test_galsim_warning():
    """Test basic usage of GalSimWarning
    """
    err = galsim.GalSimWarning("Test")
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test"
    assert isinstance(err, UserWarning)
    check_pickle(err)


@timer
def test_galsim_deprecation_warning():
    """Test basic usage of GalSimDeprecationWarning
    """
    err = galsim.GalSimDeprecationWarning("Test")
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == "Test"
    assert isinstance(err, UserWarning)
    check_pickle(err)

@timer
def test_galsim_fftsize_warning():
    """Test basic usage of GalSimFFTSizeWarning
    """
    err = galsim.GalSimFFTSizeWarning("Test", 10240)
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err).startswith("Test")
    assert isinstance(err, UserWarning)
    check_pickle(err)

@timer
def test_galsim_fft_size_error():
    """Test basic usage of GalSimFFTSizeError
    """
    err = galsim.GalSimFFTSizeError("Test FFT is too big.", 10240)
    print('str = ',str(err))
    print('repr = ',repr(err))
    assert str(err) == ("Test FFT is too big.\nThe required FFT size would be 10240 x 10240, "
                        "which requires 2.34 GB of memory.\nIf you can handle "
                        "the large FFT, you may update gsparams.maximum_fft_size.")
    assert err.size == 10240
    np.testing.assert_almost_equal(err.mem, 2.34375)
    assert isinstance(err, galsim.GalSimError)


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'Shapelet' ]

import numpy as np

from .gsobject import GSObject
from .gsparams import GSParams
from .position import PositionD
from .image import Image
from .utilities import lazy_property
from . import _galsim
from .errors import GalSimValueError, GalSimIncompatibleValuesError, GalSimNotImplementedError


class Shapelet(GSObject):
    r"""A class describing polar shapelet surface brightness profiles.

    This class describes an arbitrary profile in terms of a shapelet decomposition.  A shapelet
    decomposition is an eigenfunction decomposition of a 2-d function using the eigenfunctions
    of the 2-d quantum harmonic oscillator.  The functions are Laguerre polynomials multiplied
    by a Gaussian.  See Bernstein & Jarvis, 2002 or Massey & Refregier, 2005 for more detailed
    information about this kind of decomposition.  For this class, we follow the notation of
    Bernstein & Jarvis.

    The decomposition is described by an overall scale length, ``sigma``, and a vector of
    coefficients, ``b``.  The ``b`` vector is indexed by two values, which can be either (p,q) or
    (N,m).  In terms of the quantum solution of the 2-d harmonic oscillator, p and q are the number
    of quanta with positive and negative angular momentum (respectively).  Then, N=p+q, m=p-q.

    The 2D image is given by (in polar coordinates):

    .. math::
        I(r,\theta) = \frac{1}{\sigma^2} \sum_{pq} b_{pq} \psi_{pq}(r/\sigma, \theta)

    where :math:`\psi_{pq}` are the shapelet eigenfunctions, given by:

    .. math::
        \psi_pq(r,\theta) = \frac{(-)^q}{\sqrt{\pi}} \sqrt{\frac{q!}{p!}}
                            r^m \exp(i m \theta) \exp(-r^2/2) L_q^{(m)}(r^2)

    and :math:`L_q^{(m)}(x)` are generalized Laguerre polynomials.

    The coeffients :math:`b_{pq}` are in general complex.  However, we require that the resulting
    :math:`I(r,\theta)` be purely real, which implies that :math:`b_{pq} = b_{qp}^*`
    (where :math:`{}^*` means complex conjugate).
    This further implies that :math:`b_{pp}` (i.e. :math:`b_{pq}` with :math:`p==q`) is real.

    1. Make a blank Shapelet instance with all :math:`b_{pq} = 0.`::

        >>> shapelet = galsim.Shapelet(sigma, order)

    2. Make a Shapelet instance using a given vector for the :math:`b_{pq}` values.::

        >>> order = 2
        >>> bvec = [ 1, 0, 0, 0.2, 0.3, -0.1 ]
        >>> shapelet = galsim.Shapelet(sigma, order, bvec)

    We use the following order for the coefficients, where the subscripts are in terms of p,q.

    [ b00  Re(b10)  Im(b10)  Re(b20)  Im(b20)  b11  Re(b30)  Im(b30)  Re(b21)  Im(b21) ... ]

    i.e. we progressively increase N, and for each value of N, we start with m=N and go down to
    m=0 or 1 as appropriate.  And since m=0 is intrinsically real, it only requires one spot
    in the list.

    Parameters:
        sigma:      The scale size in the standard units (usually arcsec).
        order:      The order of the shapelet decomposition.  This is the maximum
                    N=p+q included in the decomposition.
        bvec:       The initial vector of coefficients.  [default: None, which means to use
                    all zeros]
        gsparams:   An optional `GSParams` argument. [default: None]
    """
    _req_params = { "sigma" : float, "order" : int }

    _has_hard_edges = False
    _is_axisymmetric = False
    _is_analytic_x = True
    _is_analytic_k = True

    def __init__(self, sigma, order, bvec=None, gsparams=None):
        # Make sure order and sigma are the right type:
        self._order = int(order)
        self._sigma = float(sigma)
        bvec_size = self.size(order)
        self._gsparams = GSParams.check(gsparams)

        # Make bvec if necessary
        if bvec is None:
            self._bvec = np.empty(bvec_size, dtype=float)
        else:
            if len(bvec) != bvec_size:
                raise GalSimIncompatibleValuesError(
                    "bvec is the wrong size for the provided order", bvec=bvec, order=order)
            self._bvec = np.ascontiguousarray(bvec, dtype=float)

    @lazy_property
    def _sbp(self):
        _bvec = self._bvec.__array_interface__['data'][0]
        return _galsim.SBShapelet(self._sigma, self._order, _bvec, self.gsparams._gsp)

    @classmethod
    def size(cls, order):
        """The size of the shapelet vector.
        """
        return (order+1)*(order+2)//2;

    @property
    def sigma(self):
        """The scale size, sigma.
        """
        return self._sigma

    @property
    def order(self):
        """The shapelet order.
        """
        return self._order

    @property
    def bvec(self):
        """The vector of shapelet coefficients
        """
        return self._bvec

    def getPQ(self,p,q):
        """Return the (p,q) coefficient.

        Parameters:
            p:      The p index to get.
            q:      The q index to get.

        Returns:
            a tuple (Re(b_pq), Im(b_pq))
        """
        pq = (p+q)*(p+q+1)//2 + 2*min(p,q)
        if p == q:
            return self._bvec[pq], 0
        elif p > q:
            return self._bvec[pq], self._bvec[pq+1]
        else:
            return self._bvec[pq], -self._bvec[pq+1]

    def getNM(self,N,m):
        """Return the coefficient according to N,m rather than p,q where N=p+q and m=p-q.

        Parameters:
            N:      The value of N=p+q to get.
            m:      The value of m=p-q to get.

        Returns:
            a tuple (Re(b_pq), Im(b_pq))
        """
        return self.getPQ((N+m)//2,(N-m)//2)

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, Shapelet) and
                 self.sigma == other.sigma and
                 self.order == other.order and
                 np.array_equal(self.bvec, other.bvec) and
                 self.gsparams == other.gsparams))

    def __hash__(self):
        return hash(("galsim.Shapelet", self.sigma, self.order, tuple(self.bvec), self.gsparams))

    def __repr__(self):
        return 'galsim.Shapelet(sigma=%r, order=%r, bvec=%r, gsparams=%r)'%(
                self.sigma, self.order, self.bvec, self.gsparams)

    def __str__(self):
        return 'galsim.Shapelet(sigma=%s, order=%s, bvec=%s)'%(self.sigma, self.order, self.bvec)

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_sbp',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d

    @property
    def _maxk(self):
        return self._sbp.maxK()

    @property
    def _stepk(self):
        return self._sbp.stepK()

    @property
    def _centroid(self):
        return PositionD(self._sbp.centroid())

    @property
    def _flux(self):
        return self._sbp.getFlux()

    @property
    def _positive_flux(self):
        return self._sbp.getPositiveFlux()

    @property
    def _negative_flux(self):
        return self._sbp.getNegativeFlux()

    @lazy_property
    def _flux_per_photon(self):
        return self._calculate_flux_per_photon()

    @property
    def _max_sb(self):
        return self._sbp.maxSB()

    def _xValue(self, pos):
        return self._sbp.xValue(pos._p)

    def _kValue(self, kpos):
        return self._sbp.kValue(kpos._p)

    def _drawReal(self, image, jac=None, offset=(0.,0.), flux_scaling=1.):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        dx,dy = offset
        self._sbp.draw(image._image, image.scale, _jac, dx, dy, flux_scaling)

    def _drawKImage(self, image, jac=None):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        self._sbp.drawK(image._image, image.scale, _jac)

    @classmethod
    def fit(cls, sigma, order, image, center=None, normalization='flux', gsparams=None):
        """Fit for a shapelet decomposition of a given image.

        The optional ``normalization`` parameter mirrors the parameter of the `InterpolatedImage`
        class.  The following sequence should produce drawn images that are approximate matches to
        the original image::

            >>> image = [...]
            >>> shapelet = galsim.FitShapelet(sigma, order, image, normalization='sb')
            >>> im2 = shapelet.drawImage(image=im2, scale=image.scale, method='sb')
            >>> shapelet = galsim.FitShapelet(sigma, order, image, normalization='flux')
            >>> im3 = shapelet.drawImage(image=im3, scale=image.scale, method='no_pixel')

        Then ``im2`` and ``im3`` should be as close as possible to ``image`` for the given ``sigma``
        and ``order``.  Increasing the order can improve the fit, as can having ``sigma`` match the
        natural scale size of the image.  However, it should be noted that some images are not well
        fit by a shapelet for any (reasonable) order.

        Parameters:
            sigma:          The scale size in the standard units (usually arcsec).
            order:          The order of the shapelet decomposition.  This is the maximum
                            N=p+q included in the decomposition.
            image:          The `Image` for which to fit the shapelet decomposition
            center:         The position in pixels to use for the center of the decomposition.
                            [default: image.true_center]
            normalization:  The normalization to assume for the image.
                            [default: "flux"]
            gsparams:       An optional `GSParams` argument. [default: None]

        Returns:
            the fitted Shapelet profile
        """
        if not center:
            center = image.true_center
        # convert from PositionI if necessary
        center = PositionD(center)

        if not normalization.lower() in ("flux", "f", "surface brightness", "sb"):
            raise GalSimValueError("Invalid normalization requested.", normalization,
                                   ('flux', 'f', 'surface brightneess', 'sb'))

        ret = Shapelet(sigma, order, bvec=None, gsparams=gsparams)

        if image.wcs is not None and not image.wcs._isPixelScale:
            # TODO: Add ability for ShapeletFitImage to take jacobian matrix.
            raise GalSimNotImplementedError("Sorry, cannot (yet) fit a shapelet model to an image "
                                            "with a non-trivial WCS.")

        # Make it double precision if it is not.
        image = Image(image, dtype=np.float64, copy=False)

        _bvec = ret._bvec.__array_interface__['data'][0]
        _galsim.ShapeletFitImage(ret._sigma, ret._order, _bvec,
                                 image._image, image.scale, center._p)

        if normalization.lower() == "flux" or normalization.lower() == "f":
            ret._bvec /= image.scale**2

        return ret



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import math

from .util import GetRNG
from .value import GetAllParams, CheckAllParams, RegisterValueType
from ..errors import GalSimConfigError, GalSimConfigValueError
from ..random import UniformDeviate, GaussianDeviate, PoissonDeviate, BinomialDeviate
from ..random import WeibullDeviate, GammaDeviate, Chi2Deviate, DistDeviate
from ..angle import Angle, radians
from ..position import PositionD
from ..table import LookupTable

# This file adds extra value types involving random deviates: Random, RandomGaussian,
# RandomPoisson, RandomBinomial, RandomWeibull, RandomGamma, RandomChi2, RandomDistribution,
# and RandomCircle.

def _GenerateFromRandom(config, base, value_type):
    """Return a random value drawn from a uniform distribution
    """
    rng = GetRNG(config, base)
    ud = UniformDeviate(rng)

    # Each value_type works a bit differently:
    if value_type is Angle:
        CheckAllParams(config)
        val = ud() * 2 * math.pi * radians
        #print(base['obj_num'],'Random angle = ',val)
        return val, False
    elif value_type is bool:
        opt = { 'p' : float }
        kwargs, safe = GetAllParams(config, base, opt=opt)
        p = kwargs.get('p', 0.5)
        val = ud() < p
        #print(base['obj_num'],'Random bool = ',val)
        return val, False
    else:
        ignore = [ 'default' ]
        req = { 'min' : value_type , 'max' : value_type }
        kwargs, safe = GetAllParams(config, base, req=req, ignore=ignore)

        min = kwargs['min']
        max = kwargs['max']

        if value_type is int:
            val = int(math.floor(ud() * (max-min+1))) + min
            # In case ud() == 1
            if val > max: val = max
        else:
            val = ud() * (max-min) + min

        #print(base['obj_num'],'Random = ',val)
        return val, False


def _GenerateFromRandomGaussian(config, base, value_type):
    """Return a random value drawn from a Gaussian distribution
    """
    rng = GetRNG(config, base)

    req = { 'sigma' : float }
    opt = { 'mean' : float, 'min' : float, 'max' : float }
    kwargs, safe = GetAllParams(config, base, req=req, opt=opt)

    sigma = kwargs['sigma']

    if 'gd' in base and base['current_gdsigma'] == sigma:
        # Minor subtlety here.  GaussianDeviate requires two random numbers to
        # generate a single Gaussian deviate.  But then it gets a second
        # deviate for free.  So it's more efficient to store gd than to make
        # a new one each time.  So check if we did that.
        gd = base['gd']
    else:
        # Otherwise, just go ahead and make a new one.
        gd = GaussianDeviate(rng,sigma=sigma)
        base['gd'] = gd
        base['current_gdsigma'] = sigma

    if 'min' in kwargs or 'max' in kwargs:
        # Clip at min/max.
        # However, special cases if min == mean or max == mean
        #  -- can use fabs to double the chances of falling in the range.
        mean = kwargs.get('mean',0.)
        min = kwargs.get('min',-float('inf'))
        max = kwargs.get('max',float('inf'))

        do_abs = False
        do_neg = False
        if (min >= mean) and (max > mean):
            do_abs = True
            lo = min - mean
            hi = max - mean
        elif (min < mean) and (max <= mean):
            do_abs = True
            do_neg = True
            hi = mean - min
            lo = mean - max
        else:
            lo = min - mean
            hi = max - mean

        # Emulate a do-while loop
        while True:
            val = gd()
            if do_abs: val = math.fabs(val)
            if val >= lo and val <= hi: break
        if do_neg: val = -val
        val += mean
    else:
        val = gd()
        if 'mean' in kwargs: val += kwargs['mean']

    #print(base['obj_num'],'RandomGaussian: ',val)
    return val, False

def _GenerateFromRandomPoisson(config, base, value_type):
    """Return a random value drawn from a Poisson distribution
    """
    rng = GetRNG(config, base)

    req = { 'mean' : float }
    kwargs, safe = GetAllParams(config, base, req=req)

    mean = kwargs['mean']

    dev = PoissonDeviate(rng,mean=mean)
    val = dev()

    #print(base['obj_num'],'RandomPoisson: ',val)
    return val, False

def _GenerateFromRandomBinomial(config, base, value_type):
    """Return a random value drawn from a Binomial distribution
    """
    rng = GetRNG(config, base)

    req = {}
    opt = { 'p' : float }

    # Let N be optional for bool, since N=1 is the only value that makes sense.
    if value_type is bool:
        opt['N'] = int
    else:
        req['N'] = int
    kwargs, safe = GetAllParams(config, base, req=req, opt=opt)

    N = kwargs.get('N',1)
    p = kwargs.get('p',0.5)
    if value_type is bool and N != 1:
        raise GalSimConfigValueError(
            "N must = 1 for type = RandomBinomial used in bool context", N)

    dev = BinomialDeviate(rng,N=N,p=p)
    val = dev()

    #print(base['obj_num'],'RandomBinomial: ',val)
    return val, False


def _GenerateFromRandomWeibull(config, base, value_type):
    """Return a random value drawn from a Weibull distribution
    """
    rng = GetRNG(config, base)

    req = { 'a' : float, 'b' : float }
    kwargs, safe = GetAllParams(config, base, req=req)

    a = kwargs['a']
    b = kwargs['b']
    dev = WeibullDeviate(rng,a=a,b=b)
    val = dev()

    #print(base['obj_num'],'RandomWeibull: ',val)
    return val, False


def _GenerateFromRandomGamma(config, base, value_type):
    """Return a random value drawn from a Gamma distribution
    """
    rng = GetRNG(config, base)

    req = { 'k' : float, 'theta' : float }
    kwargs, safe = GetAllParams(config, base, req=req)

    k = kwargs['k']
    theta = kwargs['theta']
    dev = GammaDeviate(rng,k=k,theta=theta)
    val = dev()

    #print(base['obj_num'],'RandomGamma: ',val)
    return val, False


def _GenerateFromRandomChi2(config, base, value_type):
    """Return a random value drawn from a Chi^2 distribution
    """
    rng = GetRNG(config, base)

    req = { 'n' : float }
    kwargs, safe = GetAllParams(config, base, req=req)

    n = kwargs['n']

    dev = Chi2Deviate(rng,n=n)
    val = dev()

    #print(base['obj_num'],'RandomChi2: ',val)
    return val, False

def _GenerateFromRandomDistribution(config, base, value_type):
    """Return a random value drawn from a user-defined probability distribution
    """
    rng = GetRNG(config, base)

    ignore = [ 'x', 'f', 'x_log', 'f_log' ]
    opt = {'function' : str, 'interpolant' : str, 'npoints' : int,
           'x_min' : float, 'x_max' : float }
    kwargs, safe = GetAllParams(config, base, opt=opt, ignore=ignore)

    # Allow the user to give x,f instead of function to define a LookupTable.
    if 'x' in config or 'f' in config:
        if 'x' not in config or 'f' not in config:
            raise GalSimConfigError(
                "Both x and f must be provided for type=RandomDistribution")
        if 'function' in kwargs:
            raise GalSimConfigError(
                "Cannot provide function with x,f for type=RandomDistribution")
        x = config['x']
        f = config['f']
        x_log = config.get('x_log', False)
        f_log = config.get('f_log', False)
        interpolant = kwargs.pop('interpolant', 'spline')
        kwargs['function'] = LookupTable(x=x, f=f, x_log=x_log, f_log=f_log,
                                         interpolant=interpolant)
    else:
        if 'function' not in kwargs:
            raise GalSimConfigError(
                "function or x,f  must be provided for type=RandomDistribution")
        if 'x_log' in config or 'f_log' in config:
            raise GalSimConfigError(
                "x_log, f_log are invalid with function for type=RandomDistribution")

    if '_distdev' not in config or config['_distdev_kwargs'] != kwargs:
        # The overhead for making a DistDeviate is large enough that we'd rather not do it every
        # time, so first check if we've already made one:
        distdev = DistDeviate(rng,**kwargs)
        config['_distdev'] = distdev
        config['_distdev_kwargs'] = kwargs
    else:
        distdev = config['_distdev']

    # Typically, the rng will change between successive calls to this, so reset the
    # seed.  (The other internal calculations don't need to be redone unless the rest of the
    # kwargs have been changed.)
    distdev.reset(rng)

    val = distdev()
    #print(base['obj_num'],'distdev = ',val)
    return val, False


def _GenerateFromRandomCircle(config, base, value_type):
    """Return a PositionD drawn from a circular top hat distribution.
    """
    rng = GetRNG(config, base)

    req = { 'radius' : float }
    opt = { 'inner_radius' : float, 'center' : PositionD }
    kwargs, safe = GetAllParams(config, base, req=req, opt=opt)
    radius = kwargs['radius']
    inner_radius = kwargs.get('inner_radius',0)

    ud = UniformDeviate(rng)
    max_rsq = radius**2
    min_rsq = inner_radius**2

    if min_rsq >= max_rsq:
        raise GalSimConfigValueError(
            "inner_radius must be less than radius (%f) for type=RandomCircle"%(radius),
            inner_radius)

    # Emulate a do-while loop
    while True:
        x = (2*ud()-1) * radius
        y = (2*ud()-1) * radius
        rsq = x**2 + y**2
        if rsq >= min_rsq and rsq <= max_rsq: break

    pos = PositionD(x,y)
    if 'center' in kwargs:
        pos += kwargs['center']

    #print(base['obj_num'],'RandomCircle: ',pos)
    return pos, False

# Register these as valid value types
RegisterValueType('Random', _GenerateFromRandom, [ float, int, bool, Angle ])
RegisterValueType('RandomGaussian', _GenerateFromRandomGaussian, [ float ])
RegisterValueType('RandomPoisson', _GenerateFromRandomPoisson, [ float, int ])
RegisterValueType('RandomBinomial', _GenerateFromRandomBinomial, [ float, int, bool ])
RegisterValueType('RandomWeibull', _GenerateFromRandomWeibull, [ float ])
RegisterValueType('RandomGamma', _GenerateFromRandomGamma, [ float ])
RegisterValueType('RandomChi2', _GenerateFromRandomChi2, [ float ])
RegisterValueType('RandomDistribution', _GenerateFromRandomDistribution, [ float ])
RegisterValueType('RandomCircle', _GenerateFromRandomCircle, [ PositionD ])



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#
"""
@file roman_detectors.py

Part of the Roman Space Telescope module.  This file includes helper routines to apply image
defects that are specific to Roman.
"""

import numpy as np
import os

from . import exptime, persistence_coefficients, nonlinearity_beta
from . import dark_current, read_noise, gain
from . import reciprocity_alpha
from . import ipc_kernel
from . import persistence_fermi_parameters

from .. import BaseDeviate, PoissonNoise, DeviateNoise, GaussianNoise, PoissonDeviate
from .. import GalSimValueError


def NLfunc(x):
    return x + nonlinearity_beta*(x**2)

def applyNonlinearity(img):
    """
    Applies the Roman nonlinearity function to the supplied image ``im``.

    For more information about nonlinearity, see the docstring for galsim.Image.applyNonlinearity.
    Unlike that routine, this one does not require any arguments, since it uses the nonlinearity
    function defined within the Roman module.

    After calling this method, the Image instance ``img`` is transformed to include the
    nonlinearity.

    Parameters:
        img:        The Image to be transformed.
    """
    img.applyNonlinearity(NLfunc=NLfunc)

def addReciprocityFailure(img, exptime=exptime):
    img.addReciprocityFailure(exp_time=exptime, alpha=reciprocity_alpha, base_flux=1.0)

# Note: Formatted doc strings don't work if put in the normal place.  Unless the function is
# actually called, the formatting statement is never executed.  So put it here instead.
addReciprocityFailure.__doc__ = """
Accounts for the reciprocity failure for the Roman directors and includes it in the original
Image ``img`` directly.

For more information about reciprocity failure, see the docstring for
galsim.Image.addReciprocityFailure.  Unlike that routine, this one does not need the parameters
for reciprocity failure to be provided, though it still takes exposure time as an optional
argument.

Parameters:
    img:            The Image to be transformed.
    exptime:        The exposure time (t) in seconds, which goes into the expression for
                    reciprocity failure given in the docstring.  If None, then the routine
                    will use the default Roman exposure time in galsim.roman.exptime.
                    [default: {exptime}]
""".format(exptime=exptime)


def applyIPC(img, edge_treatment='extend', fill_value=None):
    """
    Applies the effect of interpixel capacitance (IPC) to the Image instance.

    For more information about IPC, see the docstring for galsim.Image.applyIPC.  Unlike that
    routine, this one does not need the IPC kernel to be specified, since it uses the IPC kernel
    defined within the Roman module.

    Parameters:
        img:                The Image to be transformed.
        edge_treatment:     Specifies the method of handling edges and should be one of
                            'crop', 'extend' or 'wrap'. See galsim.Image.applyIPC docstring
                            for more information.
                            [default: 'extend']
        fill_value:         Specifies the value (including nan) to fill the edges with when
                            edge_treatment is 'crop'. If unspecified or set to 'None', the
                            original pixel values are retained at the edges. If
                            edge_treatment is not 'crop', then this is ignored.
    """
    img.applyIPC(ipc_kernel, edge_treatment=edge_treatment, fill_value=fill_value)

def applyPersistence(img, prev_exposures, method='fermi'):
    if not hasattr(prev_exposures,'__iter__'):
        raise TypeError("In roman.applyPersistence, prev_exposures must be a list of Image instances")

    if method == 'linear':

        n_exp = min(len(prev_exposures),len(persistence_coefficients))
        img.applyPersistence(prev_exposures[:n_exp], persistence_coefficients[:n_exp])

    elif method == 'fermi':

        n_exp = len(prev_exposures)
        for i in range(n_exp):
            # The slew/settle time and the reset time should be specified.
            # Now we simply assume them as 0 and take the persitence current at the mid-time of
            # exposures as the average persistence until we get more information about the
            # observation timeline.
            img.array[:,:] += fermi_linear(prev_exposures[i].array, (0.5+i)*exptime)*exptime

    else:
        raise GalSimValueError("applyPersistence only accepts 'linear' or 'fermi' methods, got",
                               method)

# Again, need to put the doc outside the function to get formatting to work.
applyPersistence.__doc__ = """
This method applies either of the two different persistence models: 'linear' and 'fermi'.
Slew between pointings and consecutive resets after illumination are not considered.

'linear' persistence model
    Applies the persistence effect to the Image instance by adding a small fraction of the
    previous exposures (up to {ncoeff}) supplied as the 'prev_exposures' argument.
    For more information about persistence, see `galsim.Image.applyPersistence`.
    Unlike that routine, this one does not need the coefficients to be specified. However,
    the list of previous {ncoeff} exposures will have to be supplied. Earlier exposures, if
    supplied, will be ignored.

'fermi' persistence model
    Applies the persistence effect to the Image instance by adding the accumulated persistence
    dark current of previous exposures supplied as the 'prev_exposures' argument.
    Unlike galsim.Image.applyPersistence, this one does not use constant coefficients but a
    fermi model plus a linear tail below half of saturation.

    For more info about the fermi model, see:

    http://www.stsci.edu/hst/wfc3/ins_performance/persistence/

Parameters:
    img:                The Image to be transformed.
    prev_exposures:     List of Image instances in the order of exposures, with the recent
                        exposure being the first element. In the linear model, the exposures
                        exceeding the limit ({ncoeff} exposures) will be ignored.
    method:             The persistence model ('linear' or 'fermi') to be applied.
                        [default: 'fermi']
""".format(ncoeff=len(persistence_coefficients))


def fermi_linear(x, t):
    """
    The fermi model for persistence: A* (x/x0)**a * (t/1000.)**(-r) / (exp( -(x-x0)/dx ) +1. )
    For influence level below the half well, the persistence is linear in x.

    Parameters:
        x:      Array of pixel influence levels in unit of electron counts.
        t:      Time (in seconds) since reset.

    Returns:
        The persistence signal of the input exposure x.
    """
    x = np.asarray(x)
    y = np.zeros_like(x)

    A, x0, dx, a, r, half_well = persistence_fermi_parameters
    ps    = A* (    x    /x0)**a * (t/1000.)**(-r)/(np.exp( -(x-x0)/dx) +1.)
    ps_hf = A* (half_well/x0)**a * (t/1000.)**(-r)/(np.exp( -(half_well-x0)/dx) +1.)

    mask1 = x > half_well
    mask2 = (x > 0.) & (x <= half_well)

    y[mask1] += ps[mask1]
    y[mask2] += ps_hf*x[mask2]/half_well

    return y

# Again, need to put the doc outside the function to get formatting to work.
def allDetectorEffects(img, prev_exposures=(), rng=None, exptime=exptime):
    # Make sure we don't have any negative values.
    img.replaceNegative(0.)

    # Add Poisson noise.
    rng = BaseDeviate(rng)
    poisson_noise = PoissonNoise(rng)
    img.addNoise(poisson_noise)

    # Quantize: have an integer number of photons in every pixel after inclusion of sky noise.
    img.quantize()

    # Reciprocity failure (use Roman routine, with the supplied exposure time).
    addReciprocityFailure(img, exptime=exptime)

    # Dark current (use exposure time).
    total_dark_current = dark_current*exptime
    dark_noise = DeviateNoise(PoissonDeviate(rng, total_dark_current))
    img.addNoise(dark_noise)

    # Persistence (use Roman H4RG-lo fermi model)
    prev_exposures = list(prev_exposures)
    applyPersistence(img, prev_exposures, method='fermi')
    # Update the 'prev_exposures' queue.
    prev_exposures = [img.copy()] + prev_exposures[:]

    # Nonlinearity (use Roman routine).
    applyNonlinearity(img)

    # IPC (use Roman routine).
    applyIPC(img)

    # Read noise.
    gn = GaussianNoise(rng, sigma=read_noise)
    img.addNoise(gn)

    # Gain.
    img /= gain

    # Quantize.
    img.quantize()

    return prev_exposures

allDetectorEffects.__doc__ = """
This utility applies all sources of noise and detector effects for Roman that are implemented
in GalSim.  In terms of noise, this includes the Poisson noise due to the signal (sky +
background), dark current, and read noise.  The detector effects that are included are
reciprocity failure, quantization, persistence, nonlinearity, and interpixel capacitance. It
also includes the necessary factors of gain.  In short, the user should be able to pass in an
Image with all sources of signal (background plus astronomical objects), and the Image will be
modified to include all subsequent steps in the image generation process for Roman that are
implemented in GalSim. However, to include the effect of persistence, the user needs to provide
a list of recent exposures (without the readout effects) and the routine
returns an updated list of recent exposures.

Parameters:
    img:            The Image to be modified.
    prev_exposures: List of Image instances in the order of exposures, with
                    the recent exposure being the first element. [default: ()]
    rng:            An optional galsim.BaseDeviate to use for the addition of noise.  If
                    None, a new one will be initialized.  [default: None]
    exptime:        The exposure time, in seconds.  If None, then the Roman default
                    exposure time will be used.  [default: {exptime}]

Returns:
    prev_exposures: Updated list of previous exposures Image instances.
""".format(exptime=exptime)




# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'Shear', '_Shear' ]

import numpy as np

from .angle import Angle, _Angle, radians
from .errors import GalSimRangeError, GalSimIncompatibleValuesError

class Shear:
    r"""A class to represent shears in a variety of ways.

    The Shear object can be initialized in a variety of ways to represent shape distortions.
    A shear is an operation that transforms a circle into an ellipse with minor-to-major axis ratio
    b/a, with position angle beta, while conserving the area (see below for a discussion of the
    implications of this choice).  Given the multiple definitions of ellipticity, we have multiple
    definitions of shear:

    reduced shear
        :math:`|g| = \frac{a - b}{a + b}`
    distortion
        :math:`|e| = \frac{a^2 - b^2}{a^2 + b^2}`
    conformal shear
        :math:`\eta = \log(b/a)`
    minor-to-major axis ratio
        :math:`q = \frac{b}{a}`

    These can be thought of as a magnitude and a real-space position angle :math:`\beta`, or as
    two components, e.g., :math:`g_1` and :math:`g_2`, with:

    .. math::

        g_1 &= |g| \cos(2 \beta) \\
        g_2 &= |g| \sin(2 \beta)

    Note: :math:`\beta` is _not_ the phase of a complex valued shear.  Rather, the complex shear is
    :math:`g_1 + i g_2 = g \exp(2 i \beta)`.  Likewise for :math:`\eta` or :math:`e`.
    The phase of the complex value is :math:`2 \beta`.

    The following are all examples of valid calls to initialize a Shear object::

        >>> s = galsim.Shear()                    # empty constructor sets ellipticity/shear to zero
        >>> s = galsim.Shear(g1=0.05, g2=0.05)
        >>> s = galsim.Shear(g1=0.05)             # assumes g2=0
        >>> s = galsim.Shear(e1=0.05, e2=0.05)
        >>> s = galsim.Shear(e2=0.05)             # assumes e1=0
        >>> s = galsim.Shear(eta1=0.07, eta2=-0.1)
        >>> s = galsim.Shear(eta=0.05, beta=45.0*galsim.degrees)
        >>> s = galsim.Shear(g=0.05, beta=0.25*numpy.pi*galsim.radians)
        >>> s = galsim.Shear(e=0.3, beta=30.0*galsim.degrees)
        >>> s = galsim.Shear(q=0.5, beta=0.0*galsim.radians)
        >>> s = galsim.Shear(0.05 + 0.03j)        # Uses the g1,g2 reduced shear definition

    There can be no mixing and matching, e.g., specifying ``g1`` and ``e2``.  It is permissible to
    only specify one of two components, with the other assumed to be zero.  If a magnitude such as
    ``e``, ``g``, ``eta``, or ``q`` is specified, then ``beta`` is also required to be specified.
    It is possible to initialize a Shear with zero reduced shear by specifying no args or kwargs,
    i.e. ``galsim.Shear()``.

    In addition, for use cases where extreme efficiency is required, you can skip all the
    normal sanity checks and branches in the regular Shear constructor by using a leading
    underscore with the complex shear ``g1 + 1j * g2``::

        >>> s = galsim._Shear(0.05 + 0.03j)  # Equivalent to galsim.Shear(g1=0.05, g2=0.03)

    Analagous to the construction options, one can access the shear in the same variety of
    definitions.

    Attributes:
        g1:         The first component of the shear in the "reduced shear" definition.
        g2:         The second component of the shear in the "reduced shear" definition.
        g:          The magnitude of the shear in the "reduced shear" definition.
        e1:         The first component of the shear in the "distortion" definition.
        e2:         The second component of the shear in the "distortion" definition.
        e:          The magnitude of the shear in the "distortion" definition.
        eta1:       The first component of the shear in the "conformal shear" definition.
        eta2:       The second component of the shear in the "conformal shear" definition.
        eta:        The magnitude of the shear in the "conformal shear" definition.
        q:          The minor-to-major axis ratio
        beta:       The position angle as an `Angle` instance
        shear:      The reduced shear as a complex number g1 + 1j * g2.

    .. note::
        Since we have defined a Shear as a transformation that preserves area, this means that it
        is not a precise description of what happens during the process of weak lensing.

        The coordinate transformation that occurs during the actual weak lensing process is such
        that if a galaxy is sheared by some :math:`(\gamma_1, \gamma_2)`, and then sheared by
        :math:`(-\gamma_1, -\gamma_2)``, it will in the end return to its original shape, but will
        have changed in area due to the magnification,

        .. math::

            \mu = \frac{1}{(1-\kappa)^2 - (\gamma_1^2 + \gamma_2^2)}

        which is not equal to 1 for non-zero shear even for convergence :math:`\kappa=0`.

        Application of a `Shear` using the `GSObject.shear` method does not include this area
        change.  To properly incorporate the effective change in area due to shear, it is necessary
        to either:

        (a) define the Shear object, use the `GSObject.shear` method, and separately use the
            `GSObject.magnify` method, or
        (b) use the `GSObject.lens` method that simultaneously magnifies and shears.
    """
    def __init__(self, *args, **kwargs):

        # There is no valid set of >2 keyword arguments, so raise an exception in this case:
        if len(kwargs) > 2:
            raise TypeError(
                "Shear constructor received >2 keyword arguments: %s"%kwargs.keys())

        if len(args) > 1:
            raise TypeError(
                "Shear constructor received >1 non-keyword arguments: %s"%str(args))

        # If a component of e, g, or eta, then require that the other component is zero if not set,
        # and don't allow specification of mixed pairs like e1 and g2.
        # Also, require a position angle if we didn't get g1/g2, e1/e2, or eta1/eta2

        # Unnamed arg must be a complex shear
        if len(args) == 1:
            self._g = args[0]
            if not isinstance(self._g, complex):
                raise TypeError("Non-keyword argument to Shear must be complex g1 + 1j * g2")

        # Empty constructor means shear == (0,0)
        elif not kwargs:
            self._g = 0j

        # g1,g2
        elif 'g1' in kwargs or 'g2' in kwargs:
            g1 = kwargs.pop('g1', 0.)
            g2 = kwargs.pop('g2', 0.)
            self._g = g1 + 1j * g2
            if abs(self._g) > 1.:
                raise GalSimRangeError("Requested shear exceeds 1.", self._g, 0., 1.)

        # e1,e2
        elif 'e1' in kwargs or 'e2' in kwargs:
            e1 = kwargs.pop('e1', 0.)
            e2 = kwargs.pop('e2', 0.)
            absesq = e1**2 + e2**2
            if absesq > 1.:
                raise GalSimRangeError("Requested distortion exceeds 1.",np.sqrt(absesq), 0., 1.)
            self._g = (e1 + 1j * e2) * self._e2g(absesq)

        # eta1,eta2
        elif 'eta1' in kwargs or 'eta2' in kwargs:
            eta1 = kwargs.pop('eta1', 0.)
            eta2 = kwargs.pop('eta2', 0.)
            eta = eta1 + 1j * eta2
            abseta = abs(eta)
            self._g = eta * self._eta2g(abseta)

        # g,beta
        elif 'g' in kwargs:
            if 'beta' not in kwargs:
                raise GalSimIncompatibleValuesError(
                    "Shear constructor requires beta when g is specified.",
                    g=kwargs['g'], beta=None)
            beta = kwargs.pop('beta')
            if not isinstance(beta, Angle):
                raise TypeError("beta must be an Angle instance.")
            g = kwargs.pop('g')
            if g > 1 or g < 0:
                raise GalSimRangeError("Requested |shear| is outside [0,1].",g, 0., 1.)
            self._g = g * np.exp(2j * beta.rad)

        # e,beta
        elif 'e' in kwargs:
            if 'beta' not in kwargs:
                raise GalSimIncompatibleValuesError(
                    "Shear constructor requires beta when e is specified.",
                    e=kwargs['e'], beta=None)
            beta = kwargs.pop('beta')
            if not isinstance(beta, Angle):
                raise TypeError("beta must be an Angle instance.")
            e = kwargs.pop('e')
            if e > 1 or e < 0:
                raise GalSimRangeError("Requested distortion is outside [0,1].", e, 0., 1.)
            self._g = self._e2g(e**2) * e * np.exp(2j * beta.rad)

        # eta,beta
        elif 'eta' in kwargs:
            if 'beta' not in kwargs:
                raise GalSimIncompatibleValuesError(
                    "Shear constructor requires beta when eta is specified.",
                    eta=kwargs['eta'], beta=None)
            beta = kwargs.pop('beta')
            if not isinstance(beta, Angle):
                raise TypeError("beta must be an Angle instance.")
            eta = kwargs.pop('eta')
            if eta < 0:
                raise GalSimRangeError("Requested eta is below 0.", eta, 0.)
            self._g = self._eta2g(eta) * eta * np.exp(2j * beta.rad)

        # q,beta
        elif 'q' in kwargs:
            if 'beta' not in kwargs:
                raise GalSimIncompatibleValuesError(
                    "Shear constructor requires beta when q is specified.",
                    q=kwargs['q'], beta=None)
            beta = kwargs.pop('beta')
            if not isinstance(beta, Angle):
                raise TypeError("beta must be an Angle instance.")
            q = kwargs.pop('q')
            if q <= 0 or q > 1:
                raise GalSimRangeError("Cannot use requested axis ratio.", q, 0., 1.)
            eta = -np.log(q)
            self._g = self._eta2g(eta) * eta * np.exp(2j * beta.rad)

        elif 'beta' in kwargs:
            raise GalSimIncompatibleValuesError(
                "beta provided to Shear constructor, but not g/e/eta/q",
                beta=kwargs['beta'], e=None, g=None, q=None, eta=None)

        # check for the case where there are 1 or 2 kwargs that are not valid ones for
        # initializing a Shear
        if kwargs:
            raise TypeError(
                "Shear constructor got unexpected extra argument(s): %s"%kwargs.keys())

    @property
    def g1(self):
        """The first component of the shear in the "reduced shear" definition.
        """
        return self._g.real

    @property
    def g2(self):
        """The second component of the shear in the "reduced shear" definition.
        """
        return self._g.imag

    @property
    def g(self):
        """The magnitude of the shear in the "reduced shear" definition.
        """
        return abs(self._g)

    @property
    def beta(self):
        """The position angle as an `Angle` instance
        """
        return _Angle(0.5 * np.angle(self._g))

    @property
    def shear(self):
        """The reduced shear as a complex number g1 + 1j * g2.
        """

        return self._g

    @property
    def e1(self):
        """The first component of the shear in the "distortion" definition.
        """
        return self._g.real * self._g2e(self.g**2)

    @property
    def e2(self):
        """The second component of the shear in the "distortion" definition.
        """
        return self._g.imag * self._g2e(self.g**2)

    @property
    def e(self):
        """The magnitude of the shear in the "distortion" definition.
        """
        return self.g * self._g2e(self.g**2)

    @property
    def esq(self):
        """The square of the magnitude of the shear in the "distortion" definition.
        """
        return self.e**2

    @property
    def eta1(self):
        """The first component of the shear in the "conformal shear" definition.
        """
        return self._g.real * self._g2eta(self.g)

    @property
    def eta2(self):
        """The second component of the shear in the "conformal shear" definition.
        """
        return self._g.imag * self._g2eta(self.g)

    @property
    def eta(self):
        """The magnitude of the shear in the "conformal shear" definition.
        """
        return self.g * self._g2eta(self.g)

    @property
    def q(self):
        """The minor-to-major axis ratio
        """
        return (1.-self.g) / (1.+self.g)

    # Helpers to convert between different conventions
    # Note: These return the scale factor by which to multiply.  Not the final value.
    def _g2e(self, absgsq):
        return 2. / (1.+absgsq)

    def _e2g(self, absesq):
        if absesq > 1.e-4:
            #return (1. - np.sqrt(1.-absesq)) / absesq
            return 1. / (1. + np.sqrt(1.-absesq))
        else:
            # Avoid numerical issues near e=0 using Taylor expansion
            return 0.5 + absesq*(0.125 + absesq*(0.0625 + absesq*0.0390625))

    def _g2eta(self, absg):
        if absg > 1.e-4:
            return 2.*np.arctanh(absg)/absg
        else:
            # This doesn't have as much trouble with accuracy, but have to avoid absg=0,
            # so might as well Taylor expand for small values.
            absgsq = absg * absg
            return 2. + absgsq*((2./3.) + absgsq*0.4)

    def _eta2g(self, abseta):
        if abseta > 1.e-4:
            return np.tanh(0.5*abseta)/abseta
        else:
            absetasq = abseta * abseta
            return 0.5 + absetasq*((-1./24.) + absetasq*(1./240.))

    # define all the various operators on Shear objects
    def __neg__(self):
        return _Shear(-self._g)

    # order of operations: shear by other._shear, then by self._shear
    def __add__(self, other):
        return _Shear((self._g + other._g) / (1. + self._g.conjugate() * other._g))

    # order of operations: shear by -other._shear, then by self._shear
    def __sub__(self, other):
        return self + (-other)

    def __eq__(self, other):
        return self is other or (isinstance(other, Shear) and self._g == other._g)
    def __ne__(self, other):
        return not self.__eq__(other)

    def getMatrix(self):
        r"""Return the matrix that tells how this shear acts on a position vector:

        If a field is sheared by some shear, s, then the position (x,y) -> (x',y')
        according to:

        .. math::

            \left( \begin{array}{c} x^\prime \\ y^\prime \end{array} \right)
            = S \left( \begin{array}{c} x \\ y \end{array} \right)

        and :math:`S` is the return value of this function ``S = shear.getMatrix()``.

        Specifically, the matrix is

        .. math::

            S = \frac{1}{\sqrt{1-g^2}}
                    \left( \begin{array}{cc} 1+g_1 & g_2 \\
                                             g_2 & 1-g_1 \end{array} \right)
        """
        return np.array([[ 1.+self.g1,  self.g2   ],
                         [  self.g2  , 1.-self.g1 ]]) / np.sqrt(1.-self.g**2)

    def rotationWith(self, other):
        r"""Return the rotation angle associated with the addition of two shears.

        The effect of two shears is not just a single net shear.  There is also a rotation
        associated with it.  This is easiest to understand in terms of the matrix representations:

        If ``shear3 = shear1 + shear2`` is a sum of two shears, and the corresponding shear
        matrices are :math:`S_1`, :math:`S_2`, and :math:`S_3`, then :math:`S_3 R = S_1 S_2`,
        where :math:`R` is a rotation matrix:

        .. math::

            R = \left( \begin{array}{cc} cos(\theta) & -sin(\theta) \\
                                         sin(\theta) &  cos(\theta) \end{array} \right)

        and :math:`\theta` is the return value (as a `galsim.Angle`) from
        ``shear1.rotationWith(shear2)``.
        """
        # Save a little time by only working on the first column.
        S3 = self.getMatrix().dot(other.getMatrix()[:,:1])
        R = (-(self + other)).getMatrix().dot(S3)
        theta = np.arctan2(R[1,0], R[0,0])
        return theta * radians

    def __repr__(self):
        return 'galsim.Shear(%r)'%(self.shear)

    def __str__(self):
        return 'galsim.Shear(g1=%s,g2=%s)'%(self.g1,self.g2)

    def __hash__(self): return hash(self._g)

def _Shear(shear):
    """Equivalent to ``galsim.Shear(shear)``, but without the overhead of the normal sanity checks
    and other options for how to specify the shear.

    Parameters:
        shear:      The complex shear g1 + 1j * g2.

    Returns:
        a `galsim.Shear` instance
    """
    ret = Shear.__new__(Shear)
    ret._g = shear
    return ret



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import logging
import copy
import sys
import time
import json
from io import StringIO
import cProfile, pstats
from collections import OrderedDict
from multiprocessing.managers import BaseManager
from multiprocessing import current_process, get_context, cpu_count
from queue import Empty
import traceback

from ..utilities import SimpleGenerator, single_threaded
from ..random import BaseDeviate
from ..errors import GalSimConfigError, GalSimConfigValueError, GalSimError, galsim_warn

max_queue_size = 32767  # This is where multiprocessing.Queue starts to have trouble.
                        # We make it a settable parameter here really for unit testing.
                        # I don't think there is any reason for end users to want to set this.

def MergeConfig(config1, config2, logger=None):
    """
    Merge config2 into config1 such that it has all the information from either config1 or
    config2 including places where both input dicts have some of a field defined.

    e.g. If config1 has image.pixel_scale, and config2 has image.noise, then the returned dict
    will have both.

    For real conflicts (the same value in both cases), config1's value takes precedence
    """
    logger = LoggerWrapper(logger)
    for (key, value) in config2.items():
        if not key in config1:
            # If this key isn't in config1 yet, just add it
            config1[key] = copy.deepcopy(value)
        elif isinstance(value,dict) and isinstance(config1[key],dict):
            # If they both have a key, first check if the values are dicts
            # If they are, just recurse this process and merge those dicts.
            MergeConfig(config1[key],value,logger)
        else:
            # Otherwise config1 takes precedence
            logger.info("Not merging key %s from the base config, since the later "
                        "one takes precedence",key)
            pass

def ReadYaml(config_file):
    """Read in a YAML configuration file and return the corresponding dicts.

    A YAML file is allowed to define several dicts using multiple documents. The GalSim parser
    treats this as a set of multiple jobs to be done.  The first document is taken to be a "base"
    dict that has common definitions for all the jobs.  Then each subsequent document has the
    (usually small) modifications to the base dict for each job.  See demo6.yaml, demo8.yaml and
    demo9.yaml in the GalSim/examples directory for example usage.

    The return value will be a list of dicts, one dict for each job to be done.

    Parameters:
        config_file:    The name of the configuration file to read.

    Returns:
        list of config dicts
    """
    import yaml

    with open(config_file) as f:
        all_config = [ c for c in yaml.load_all(f.read(), yaml.SafeLoader) ]

    # If there is only 1 yaml document, then it is of course used for the configuration.
    # If there are multiple yaml documents, then the first one defines a common starting
    # point for the later documents.
    # So the configurations are taken to be:
    #   all_config[0] + all_config[1]
    #   all_config[0] + all_config[2]
    #   all_config[0] + all_config[3]
    #   ...
    # See demo6.yaml and demo8.yaml in the examples directory for examples of this feature.

    if len(all_config) > 1:
        # Break off the first one if more than one:
        base_config = all_config[0]
        all_config = all_config[1:]
        # And merge it into each of the other dicts
        for c in all_config:
            MergeConfig(c, base_config)

    return all_config


def ReadJson(config_file):
    """Read in a JSON configuration file and return the corresponding dicts.

    A JSON file only defines a single dict.  However to be parallel to the functionality of
    ReadYaml, the output is a list with a single item, which is the dict defined by the JSON file.

    Parameters:
        config_file:    The name of the configuration file to read.

    Returns:
        [config_dict]
    """

    with open(config_file) as f:
        # cf. http://stackoverflow.com/questions/6921699/can-i-get-json-to-load-into-an-ordereddict-in-python
        config = json.load(f, object_pairs_hook=OrderedDict)

    # JSON files only ever define a single job, but we need to return a list with this one item.
    return [config]


def ConvertNones(config):
    """Convert any items whose value is 'None' to None.

    To allow some parameters to be set to None in the config dict (e.g. in a list, where only
    some values need to be None), we convert all values == 'None' to None.

    Parameters:
        config:     The config dict to process
    """
    if isinstance(config, dict):
        keys = config.keys()
    else:
        keys = range(len(config))

    for key in keys:
        # Recurse to lower levels, if any
        if isinstance(config[key],(list,dict)):
            ConvertNones(config[key])

        # Convert any Nones at this level
        elif config[key] == 'None':
            config[key] = None


def RemoveCurrent(config, keep_safe=False, type=None, index_key=None):
    """
    Remove any "current" values stored in the config dict at any level.

    Parameters:
        config:     The configuration dict.
        keep_safe:  Should current values that are marked as safe be preserved?
                    [default: False]
        type:       If provided, only clear the current value of objects that use this
                    particular type.  [default: None, which means to remove current values
                    of all types.]
        index_key:  If provided, only clear the current value of objects that use this
                    index_key (or start with this index_key, so obj_num also does
                    obj_num_in_file).  [default: None]
    """
    # End recursion if this is not a dict.
    if not isinstance(config,dict): return

    # Recurse to lower levels, if any
    force = False  # If lower levels removed anything, then force removal at this level as well.
    for key in config:
        if isinstance(key, str) and key[0] == '_':
            continue  # These are our own implementation details, not the normal dict.
        if isinstance(config[key],list):
            for item in config[key]:
                force = RemoveCurrent(item, keep_safe, type, index_key) or force
        else:
            force = RemoveCurrent(config[key], keep_safe, type, index_key) or force
    if force:
        keep_safe = False
        type = None
        index_key = None

    # Delete the current_val at this level, if any
    if 'current' in config:
        cval, csafe, ctype, cindex, cindex_key = config['current']
        if (not (keep_safe and csafe)
                and (type is None or ('type' in config and config['type'] == type))
                and (index_key is None or cindex_key.startswith(index_key))):
            del config['current']
            return True
    return force

def CopyConfig(config):
    """
    If you want to use a config dict for multiprocessing, you need to deep copy
    the gal, psf, and pix fields, since they cache values that are not picklable.
    If you don't do the deep copy, then python balks when trying to send the updated
    config dict back to the root process.  We do this a few different times, so encapsulate
    the copy semantics once here.

    Parameters:
        config:     The configuration dict to copy.

    Returns:
        a deep copy of the config dict.
    """
    from .process import top_level_fields, rng_fields

    # Start with a shallow copy of the regular things that aren't one of our special fields.
    # Make sure not to copy the input_manager if there is one, since using it isn't thread safe.
    galsim_fields = top_level_fields + rng_fields
    config1 = { k:v for k,v in config.items() if k not in galsim_fields and k != '_input_manager' }

    def recursive_copy(d):
        # copy the given dict, skipping any current leading underscore fields.
        if isinstance(d, dict):
            # Use type(d) rather than dict comprehension, since in python 2.7, 3.5, this may
            # be an OrderedDict, and we want to preserve that.
            return type(d)([(k,recursive_copy(v)) for k,v in d.items() if not k.startswith('_')])
        elif isinstance(d, list):
            return [recursive_copy(v) for v in d]
        else:
            return d

    # Now copy all the regular config fields to make sure things like current don't
    # get clobbered by two processes writing to the same dict.  Keep any current items
    # that are already here, since they should be safe, but remove leading underscore items.
    for field in top_level_fields:
        if field in config:
            config1[field] = recursive_copy(config[field])

    # And deep copy any rng fields:
    for field in rng_fields:
        if field in config:
            config1[field] = copy.deepcopy(config[field])

    return config1

class SafeManager(BaseManager):
    """There are a few places we need a Manager object.  This one uses the 'fork' context,
    rather than whatever the default is on your system (which may be fork or may be spawn).

    Starting in python 3.8, the spawn context started becoming more popular.  It's supposed to
    be safer, but it wants to pickle a lot of things that aren't picklable, so it doesn't work.
    Using this as the base class instead, should work.  And doing it it one place means that we
    only have one place to change this is there is a different strategy that works better.
    """
    def __init__(self):
        super(SafeManager, self).__init__(ctx=get_context('fork'))


def GetLoggerProxy(logger):
    """Make a proxy for the given logger that can be passed into multiprocessing Processes
    and used safely.

    Parameters:
        logger:      The logger to make a copy of

    Returns:
        a proxy for the given logger
    """
    if logger:
        class LoggerManager(SafeManager): pass
        logger_generator = SimpleGenerator(logger)
        LoggerManager.register('logger', callable = logger_generator)
        logger_manager = LoggerManager()
        with single_threaded():
            logger_manager.start()
        logger_proxy = logger_manager.logger()
    else:
        logger_proxy = None
    return logger_proxy

class LoggerWrapper:
    """A wrap around a Logger object that checks whether a debug or info or warn call will
    actually produce any output before calling the functions.

    This seems like a gratuitous wrapper, and it is if the object being wrapped is a real
    Logger object.  However, we use it to wrap proxy objects (returned from GetLoggerProxy)
    that would otherwise send the arguments of logger.debug(...) calls through a multiprocessing
    pipe before (typically) being ignored.  Here, we check whether the call will actually
    produce any output before calling the functions.

    Parameters:
        logger:     The logger object to wrap.
    """
    def __init__(self, logger):
        if isinstance(logger,LoggerWrapper):
            self.logger = logger.logger
            self.level = logger.level
        else:
            self.logger = logger
            # When multiprocessing, it is faster to check if the level is enabled locally, rather
            # than communicating over the pipe to ask the base logger if it isEnabledFor a given
            # level.
            # If the logger is None, use more than the top logging level (CRITICAL).
            self.level = self.logger.getEffectiveLevel() if self.logger else logging.CRITICAL+1

    def getEffectiveLevel(self):
        return self.level

    def __bool__(self):
        return self.logger is not None
    __nonzero__ = __bool__

    def debug(self, *args, **kwargs):
        if self.logger and self.isEnabledFor(logging.DEBUG):
            self.logger.debug(*args, **kwargs)

    def info(self, *args, **kwargs):
        if self.logger and self.isEnabledFor(logging.INFO):
            self.logger.info(*args, **kwargs)

    def warning(self, *args, **kwargs):
        if self.logger and self.isEnabledFor(logging.WARNING):
            self.logger.warning(*args, **kwargs)

    def error(self, *args, **kwargs):
        if self.logger and self.isEnabledFor(logging.ERROR):
            self.logger.error(*args, **kwargs)

    def critical(self, *args, **kwargs):
        if self.logger and self.isEnabledFor(logging.CRITICAL):
            self.logger.critical(*args, **kwargs)

    def log(self, level, *args, **kwargs):
        if self.logger and self.isEnabledFor(level):
            self.logger.log(level, *args, **kwargs)

    def isEnabledFor(self, level):
        return level >= self.level


def UpdateNProc(nproc, ntot, config, logger=None):
    """Update nproc

    - If nproc < 0, set nproc to ncpu
    - Make sure nproc <= ntot

    Parameters:
        nproc:      The nominal number of processes from the config dict
        ntot:       The total number of files/images/stamps to do, so the maximum number of
                    processes that would make sense.
        config:     The configuration dict to copy.
        logger:     If given, a logger object to log progress. [default: None]

    Returns:
        the number of processes to use.
    """
    logger = LoggerWrapper(logger)
    # First if nproc < 0, update based on ncpu
    if nproc <= 0:
        # Try to figure out a good number of processes to use
        try:
            nproc = cpu_count()
            logger.debug("ncpu = %d.",nproc)
        except Exception as e:
            logger.warning("nproc <= 0, but unable to determine number of cpus.")
            logger.warning("Caught error: %s",e)
            logger.warning("Using single process")
            nproc = 1

    # Second, make sure we aren't already in a multiprocessing mode
    if nproc > 1 and 'current_nproc' in config:
        logger.debug("Already multiprocessing.  Ignoring image.nproc")
        nproc = 1

    # Third, pypy multiprocessing in Python 3.6 seems to be flaky with multiprocessing,
    # so don't use multiple processes on that sysem.
    if nproc != 1 and 'PyPy' in sys.version and '3.6' in sys.version:  # pragma: no cover
        logger.warning("PyPy 3.6 multiprocessing is unreliable.  Ignoring nproc.")
        nproc = 1

    # Finally, don't try to use more processes than jobs.  It wouldn't fail or anything.
    # It just looks bad to have 3 images processed with 8 processes or something like that.
    if nproc > ntot:
        logger.debug("There are only %d jobs to do.  Reducing nproc to %d."%(ntot,ntot))
        nproc = ntot
    return nproc


def ParseRandomSeed(config, param_name, base, seed_offset):
    """Parse the random_seed field, converting an integer (initial) seed value into a
    Sequence.
    """
    from .value import ParseValue

    # Normally, random_seed parameter is just a number, which really means to use that seed
    # start a sequential sequence of seeds for each object.
    # However, we allow for random_seed to be a gettable parameter, so for the
    # normal case, we just convert it into a Sequence.
    # We only do this for the first item in the list though.  Any items after the first
    # are left as they are, so the user can do something more complicated if they want.
    if param_name in ['random_seed', 0] and not (
        isinstance(config[param_name], dict) and config[param_name].get('default', False)):
            if (isinstance(config[param_name], dict)
                and config[param_name].get('type',None) == 'Sequence'):
                    # Not really a deprecation, but similar.
                    # Warn users of the change in behavior.
                    galsim_warn('You have a custom Sequence as the (first) random_seed. '
                                'As of GalSim version 2.6, this will be parsed as an integer '
                                'and converted to a new sequence indexed by obj_num, '
                                'rather than whatever sequence you have specified. '
                                'You probably want to put your custom random_seed sequence '
                                'as the second item in the random_seed list and use rng_num=1.')

            first = ParseValue(config, param_name, base, int)[0]
            seed_rng = BaseDeviate(first)
            new_first = seed_rng.raw()
            # Note: This new "first" seed is actually the seed value to use for anything at file or
            # image scope using the obj_num of the first object in the file or image.
            # Seeds for objects will start at 1 more than this.
            config[param_name] = {
                    'type' : 'Sequence',
                    'index_key' : 'obj_num',
                    'first' : new_first,
                    # This is out indicator that the dict has been converted from an initial
                    # integer value to generate the sequence.  After the first time, we don't
                    # want to redo this step.
                    'default' : True
            }

    index_key = base['index_key']
    if index_key == 'obj_num':
        # The normal case
        seed = ParseValue(config, param_name, base, int)[0]
    else:
        # If we are setting either the file_num or image_num rng, we need to be careful.
        base['index_key'] = 'obj_num'
        seed = ParseValue(config, param_name, base, int)[0]
        base['index_key'] = index_key
    seed += seed_offset

    # Normally, the file_num rng and the image_num rng can be the same.  So if seed
    # comes out the same as what we already built, then just use the existing file_num_rng.
    if index_key == 'image_num' and seed == base.get('file_num_seed',None):
        rng = base['file_num_rng']
    else:
        rng = BaseDeviate(seed)

    return seed, rng

def PropagateIndexKeyRNGNum(config, index_key=None, rng_num=None, rng_index_key=None, key=None):
    """Propagate any index_key or rng_num specification in a dict to all sub-fields
    """
    if isinstance(config, list):
        for item in config:
            PropagateIndexKeyRNGNum(item, index_key, rng_num)
        return

    if not isinstance(config, dict): return

    if 'index_key' in config:
        index_key = config['index_key']
    elif index_key is not None:
        config['index_key'] = index_key

    if 'rng_num' in config:
        rng_num = config['rng_num']
    elif rng_num is not None:
        config['rng_num'] = rng_num

    if 'rng_index_key' in config:
        rng_index_key = config['rng_index_key']
    elif rng_index_key is not None:
        config['rng_index_key'] = rng_index_key

    if key is None:
        for key, field in config.items():
            if isinstance(key, str) and key[0] == '_': continue
            PropagateIndexKeyRNGNum(field, index_key, rng_num, rng_index_key)
    else:
        PropagateIndexKeyRNGNum(config[key], index_key, rng_num, rng_index_key)


def SetupConfigRNG(config, seed_offset=0, logger=None):
    """Set up the RNG in the config dict.

    - Setup config['image']['random_seed'] if necessary
    - Set config['rng'] and other related values based on appropriate random_seed

    Parameters:
        config:         The configuration dict.
        seed_offset:    An offset to use relative to what config['image']['random_seed'] gives.
                        [default: 0]
        logger:         If given, a logger object to log progress. [default: None]

    Returns:
        the seed used to initialize the RNG.
    """
    from .process import rng_fields

    logger = LoggerWrapper(logger)

    # If we are starting a new file, clear out the existing rngs.
    index_key = config['index_key']
    if index_key == 'file_num':
        for key in rng_fields + ['seed', 'obj_num_seed', 'image_num_seed', 'file_num_seed']:
            config.pop(key, None)

    # This can be present for efficiency, since GaussianDeviates produce two values at a time,
    # so it is more efficient to not create a new GaussianDeviate object each time.
    # But if so, we need to remove it now.
    config.pop('gd', None)

    # All fields have a default index_key that they would normally use as well as a default
    # rng_num = 0 if the random_seed is a list.  But we allow users to specify alternate values
    # of index_key and/or rng_num at any level in the dict.  We want that specification to
    # propagate down to lower levels from that point forward.  The easiest way to do so is to
    # explicitly run through the dict once and propagate any index_key and/or rng_num fields
    # to all sub-fields.
    if not config.get('_propagated_index_key_rng_num',False):
        logger.debug('Propagating any index_key or rng_num specifications')
        for field in config.values():
            PropagateIndexKeyRNGNum(field)
        config['_propagated_index_key_rng_num'] = True

    if 'image' not in config or 'random_seed' not in config['image']:
        logger.debug('obj %d: No random_seed specified.  Using /dev/urandom',
                     config.get('obj_num',0))
        config['seed'] = 0
        rng = BaseDeviate()
        config['rng'] = rng
        config[index_key + '_rng'] = rng
        return 0

    image = config['image']
    # Normally, there is just one random_seed and it is just an integer.  However, sometimes
    # it is useful to have 2 (or more) rngs going with different seed sequences.  To enable this,
    # image.random_seed is allowed to be a list, each item of which may be either an integer or
    # a dict defining an integer sequence.  If it is a list, we parse each item separately
    # and then put the combined results into config['rng'] as a list.
    if isinstance(image['random_seed'], list):
        lst = image['random_seed']
        logger.debug('random_seed = %s',CleanConfig(lst))
        logger.debug('seed_offset = %s',seed_offset)
        seeds = []
        rngs = []
        for i in range(len(lst)):
            seed, rng = ParseRandomSeed(lst, i, config, seed_offset)
            logger.debug('seed %d = %s',i,seed)
            seeds.append(seed)
            rngs.append(rng)
            if i == 0:
                # Helpful to get this done right away, because later seeds might be based
                # on a random number that uses the first rng.
                # cf. test_eval_full_word in test_config_output.py.
                config['seed'] = seed
                config['rng'] = rng
                config[index_key + '_seed'] = seed
                config[index_key + '_rng'] = rng
        config[index_key + '_rngs'] = rngs
        logger.debug('obj %d: random_seed is a list. Initializing rngs with seeds %s',
                     config.get('obj_num',0), seeds)
        return seeds[0]
    else:
        seed, rng = ParseRandomSeed(image, 'random_seed', config, seed_offset)
        config['seed'] = seed
        config['rng'] = rng
        config[index_key + '_seed'] = seed
        config[index_key + '_rng'] = rng
        logger.debug('obj %d: Initializing rng with seed %s', config.get('obj_num',0), seed)
        return seed


def ParseExtendedKey(config, key):
    """Traverse all but the last item in an extended key and return the resulting config, key.

    If key is an extended key like gal.items.0.ellip.e, then this will return the tuple.
    (config['gal']['items'][0]['ellip'], 'e').

    If key is a regular string, then is just returns the original (config, key).

    Parameters:
        config:     The configuration dict.
        key:        The possibly extended key.

    Returns:
        the equivalent (config, key) where key is now a regular non-extended key.
    """
    # This is basically identical to the code for Dict.get(key) in catalog.py.
    chain = key.split('.')
    while True:
        d = config
        k = chain.pop(0)
        try: k = int(k)
        except ValueError: pass
        if len(chain) == 0: break
        try:
            config = d[k]
        except (TypeError, KeyError):
            # TypeError for the case where d is a float or Position2D, so d[k] is invalid.
            # KeyError for the case where d is a dict, but k is not a valid key.
            raise GalSimConfigError(
                "Unable to parse extended key %s.  Field %s is invalid."%(key,k)) from None
    return d, k

def GetFromConfig(config, key):
    """Get the value for the (possibly extended) key from a config dict.

    If key is a simple string, then this is equivalent to config[key].
    However, key is allowed to be a chain of keys such as 'gal.items.0.ellip.e', in which
    case this function will return config['gal']['items'][0]['ellip']['e'].

    Parameters:
        config:     The configuration dict.
        key:        The possibly extended key.

    Returns:
        the value of that key from the config.
    """
    d, k = ParseExtendedKey(config, key)
    try:
        value = d[k]
    except Exception as e:
        raise GalSimConfigError(
            "Unable to parse extended key %s.  Field %s is invalid."%(key,k)) from None
    return value

def SetInConfig(config, key, value, logger=None):
    """Set the value of a (possibly extended) key in a config dict.

    If key is a simple string, then this is equivalent to config[key] = value.
    However, key is allowed to be a chain of keys such as 'gal.items.0.ellip.e', in which
    case this function will set config['gal']['items'][0]['ellip']['e'] = value.

    Parameters:
        config:     The configuration dict.
        key:        The possibly extended key.

    Returns:
        the value of that key from the config.
    """
    d, k = ParseExtendedKey(config, key)
    if value == '':
        # This means remove it, if it is there.
        try:
            d.pop(k,None)
        except TypeError:
            # Then d was really a list.
            # This has some potentially counter-intuitive consequences, so let the user
            # know about them.
            if logger:
                logger.warning("Warning: Removing item %d from %s. "%(k,key[:key.rfind('.')]) +
                               "Any further adjustments to this list must use the new "+
                               "list index values, not the original indices.")
            del d[k]
    else:
        try:
            d[k] = value
        except Exception as e:
            raise GalSimConfigError(
                "Unable to parse extended key %s.  Field %s is invalid."%(key,k)) from None


def UpdateConfig(config, new_params, logger=None):
    """Update the given config dict with additional parameters/values.

    Parameters:
        config:         The configuration dict to update.
        new_params:     A dict of parameters to update.  The keys of this dict may be
                        chained field names, such as gal.first.dilate, which will be
                        parsed to update config['gal']['first']['dilate'].
    """
    for key, value in new_params.items():
        SetInConfig(config, key, value, logger)


def MultiProcess(nproc, config, job_func, tasks, item, logger=None, timeout=900,
                 done_func=None, except_func=None, except_abort=True):
    """A helper function for performing a task using multiprocessing.

    A note about the nomenclature here.  We use the term "job" to mean the job of building a single
    file or image or stamp.  The output of each job is gathered into the list of results that
    is returned.  A task is a collection of one or more jobs that are all done by the same
    processor.  For simple cases, each task is just a single job, but for things like a Ring
    test, the task needs to have the jobs for a full ring.

    The tasks argument is a list of tasks.
    Each task in that list is a list of jobs.
    Each job is a tuple consisting of (kwargs, k), where kwargs is the dict of kwargs to pass to
    the job_func and k is the index of this job in the full list of jobs.

    Parameters:
        nproc:          How many processes to use.
        config:         The configuration dict.
        job_func:       The function to run for each job.  It will be called as::

                            result = job_func(**kwargs)

                        where kwargs is from one of the jobs in the task list.
        tasks:          A list of tasks to run.  Each task is a list of jobs, each of which is
                        a tuple (kwargs, k).
        item:           A string indicating what is being worked on.
        logger:         If given, a logger object to log progress. [default: None]
        timeout:        How many seconds to allow for each task before timing out. [default: 900]
        done_func:      A function to run upon completion of each job.  It will be called as::

                            done_func(logger, proc, k, result, t)

                        where proc is the process name, k is the index of the job, result is
                        the return value of that job, and t is the time taken. [default: None]
        except_func:    A function to run if an exception is encountered.  It will be called as::

                            except_func(logger, proc, k, ex, tr)

                        where proc is the process name, k is the index of the job that failed,
                        ex is the exception caught, and tr is the traceback. [default: None]
        except_abort:   Whether an exception should abort the rest of the processing.
                        If False, then the returned results list will not include anything
                        for the jobs that failed.  [default: True]

    Returns:
        a list of the outputs from job_func for each job
    """
    from .input import worker_init_fns, worker_initargs_fns

    # The worker function will be run once in each process.
    # It pulls tasks off the task_queue, runs them, and puts the results onto the results_queue
    # to send them back to the main process.
    # The *tasks* can be made up of more than one *job*.  Each job involves calling job_func
    # with the kwargs from the list of jobs.
    # Each job also carries with it its index in the original list of all jobs.
    def worker(task_queue, results_queue, config, logger, initializers, initargs):
        proc = current_process().name

        for init, args in zip(initializers, initargs):
            init(*args)

        # The logger object passed in here is a proxy object.  This means that all the arguments
        # to any logging commands are passed through the pipe to the real Logger object on the
        # other end of the pipe.  This tends to produce a lot of unnecessary communication, since
        # most of those commands don't actually produce any output (e.g. logger.debug(..) commands
        # when the logging level is not DEBUG).  So it is helpful to wrap this object in a
        # LoggerWrapper that checks whether it is worth sending the arguments back to the original
        # Logger before calling the functions.
        logger = LoggerWrapper(logger)

        if 'profile' in config and config['profile']:
            pr = cProfile.Profile()
            pr.enable()
        else:
            pr = None

        for task in iter(task_queue.get, 'STOP'):
            try :
                logger.debug('%s: Received job to do %d %ss, starting with %s',
                             proc,len(task),item,task[0][1])
                for kwargs, k in task:
                    t1 = time.time()
                    kwargs['config'] = config
                    kwargs['logger'] = logger
                    result = job_func(**kwargs)
                    t2 = time.time()
                    results_queue.put( (result, k, t2-t1, proc) )
            except Exception as e:
                tr = traceback.format_exc()
                logger.debug('%s: Caught exception: %s\n%s',proc,str(e),tr)
                results_queue.put( (e, k, tr, proc) )
        logger.debug('%s: Received STOP', proc)
        if pr is not None:
            pr.disable()
            pr.dump_stats(config.get('root', 'galsim') + '-' + str(proc) + '.pstats')
            s = StringIO()
            sortby = 'time'  # Note: This is now called tottime, but time seems to be a valid
                             # alias for this that is backwards compatible to older versions
                             # of pstats.
            ps = pstats.Stats(pr, stream=s).sort_stats(sortby).reverse_order()
            ps.print_stats()
            logger.error("*** Start profile for %s ***\n%s\n*** End profile for %s ***",
                         proc,s.getvalue(),proc)

    njobs = sum([len(task) for task in tasks])

    if nproc > 1:
        logger.warning("Using %d processes for %s processing",nproc,item)

        ctx = get_context('fork')
        Process = ctx.Process
        Queue = ctx.Queue

        # Temporarily mark that we are multiprocessing, so we know not to start another
        # round of multiprocessing later.
        config['current_nproc'] = nproc

        if 'profile' in config and config['profile']:
            logger.info("Starting separate profiling for each of the %d processes.",nproc)

        # The logger is not picklable, so we need to make a proxy for it so all the
        # processes can emit logging information safely.
        logger_proxy = GetLoggerProxy(logger)

        # We need to set the number of OpenMP threads to 1 during multiprocessing, otherwise
        # it may spawn e.g. 64 threads in each of 64 processes, which tends to be bad.
        with single_threaded():

            # Send the tasks to the task_queue.
            ntasks = len(tasks)
            if ntasks > max_queue_size:
                logger.info("len(tasks) = %d is more than max_queue_size = %d",
                            len(tasks),max_queue_size)
                njoin = (ntasks-1) // max_queue_size + 1
                new_size = (ntasks-1) // njoin + 1
                tasks = [
                    [job for task in tasks[njoin*k : min(njoin*(k+1),ntasks)] for job in task]
                    for k in range(new_size)
                ]
                ntasks = len(tasks)
                logger.info("Joined in groups of %d: new len(tasks) = %d",njoin,ntasks)
            task_queue = Queue(ntasks)
            for task in tasks:
                task_queue.put(task)

            # Run the tasks.
            # Each Process command starts up a parallel process that will keep checking the queue
            # for a new task. If there is one there, it grabs it and does it. If not, it waits
            # until there is one to grab. When it finds a 'STOP', it shuts down.
            results_queue = Queue(ntasks)
            p_list = []
            for j in range(nproc):
                # The process name is actually the default name that Process would generate on its
                # own for the first time we do this. But after that, if we start another round of
                # multiprocessing, then it just keeps incrementing the numbers, rather than starting
                # over at Process-1.  As far as I can tell, it's not actually spawning more
                # processes, so for the sake of the logging output, we name the processes explicitly.
                initializers = worker_init_fns
                initargs = [initargs_fn() for initargs_fn in worker_initargs_fns]
                p = Process(target=worker, args=(task_queue, results_queue, config, logger_proxy,
                                                 initializers, initargs),
                            name='Process-%d'%(j+1))
                p.start()
                p_list.append(p)

            raise_error = None

            try:
                # In the meanwhile, the main process keeps going.  We pull each set of images off
                # of the results_queue and put them in the appropriate place in the lists.
                # This loop is happening while the other processes are still working on their tasks.
                results = [ None for k in range(njobs) ]
                for kk in range(njobs):
                    try:
                        res, k, t, proc = results_queue.get(timeout=timeout)
                    except Empty:
                        # If timeout, then a queue.Empty is raised, which is pretty uninformative.
                        # So reraise with a better description of what happened.
                        logger.error("Multiprocessing timed out waiting for a task to finish.")
                        logger.error("If you expect very long jobs, consider raising the value"
                                     "of timeout.  Currently set to %s",timeout)
                        raise GalSimError("Multiprocessing timed out") from None
                    if isinstance(res, Exception):
                        # res is really the exception, e
                        # t is really the traceback
                        # k is the index for the job that failed
                        if except_func is not None:  # pragma: no branch
                            except_func(logger, proc, k, res, t)
                        if except_abort:
                            for j in range(nproc):
                                p_list[j].terminate()
                            raise_error = res
                            break
                    else:
                        # The normal case
                        if done_func is not None:  # pragma: no branch
                            done_func(logger, proc, k, res, t)
                        results[k] = res

            except BaseException as e:
                # I'm actually not sure how to make this happen.  We do a good job of catching
                # all the normal kinds of exceptions that might occur and dealing with them cleanly.
                # However, if something happens that we didn't anticipate, this should make an
                # attempt to clean up the task queue and worker before raising the error further.
                logger.error("Caught a fatal exception during multiprocessing:\n%r",e)
                logger.error("%s",traceback.format_exc())
                # Clear any unclaimed jobs that are still in the queue
                while not task_queue.empty():
                    try:
                        # If something else gets between the empty() check and this get,
                        # it will raise an Empty exception.  Ignore it.
                        task_queue.get_nowait()
                    except Exception:  # pragma: no cover
                        pass
                # And terminate any jobs that might still be running.
                for j in range(nproc):
                    p_list[j].terminate()
                raise_error = e

            finally:
                # Stop the processes
                # Once you are done with the processes, putting nproc 'STOP's will stop them all.
                # This is important, because the program will keep running as long as there are
                # running processes, even if the main process gets to the end.  So you do want to
                # make sure to add those 'STOP's at some point!
                for j in range(nproc):
                    try:
                        # Sometimes this one will hang if we timed out above.  Not sure why, but
                        # the nowait seems to fix it.  And if there was a reason for put() to
                        # hang, then put_nowait might raise an exception.  Ignore it.
                        task_queue.put_nowait('STOP')
                    except Exception:  # pragma: no cover
                        pass
                for j in range(nproc):
                    p_list[j].join()
                task_queue.close()

        del config['current_nproc']

        if raise_error is not None:
            raise raise_error

    else : # nproc == 1
        results = [ None ] * njobs
        for task in tasks:
            for kwargs, k in task:
                try:
                    t1 = time.time()
                    kwargs['config'] = config
                    kwargs['logger'] = logger
                    result = job_func(**kwargs)
                    t2 = time.time()
                    if done_func is not None:  # pragma: no branch
                        done_func(logger, None, k, result, t2-t1)
                    results[k] = result
                except Exception as e:
                    tr = traceback.format_exc()
                    if except_func is not None: # pragma: no branch
                        except_func(logger, None, k, e, tr)
                    if except_abort:
                        raise

    # If there are any failures, then there will still be some Nones in the results list.
    # Remove them.
    results = [ r for r in results if r is not None ]

    return results


valid_index_keys = [ 'obj_num_in_file', 'obj_num', 'image_num', 'file_num' ]

def GetIndex(config, base, is_sequence=False):
    """Return the index to use for the current object or parameter and the index_key.

    First check for an explicit index_key value given by the user.
    Then if base[index_key] is other than obj_num, use that.
    Finally, if this is a sequence, default to 'obj_num_in_file', otherwise 'obj_num'.

    Returns:
        index, index_key
    """
    if 'index_key' in config:
        index_key = config['index_key']
        if index_key not in valid_index_keys:
            raise GalSimConfigValueError("Invalid index_key.", index_key, valid_index_keys)
    else:
        index_key = base.get('index_key','obj_num')
        if index_key == 'obj_num' and is_sequence:
            index_key = 'obj_num_in_file'

    if index_key == 'obj_num_in_file':
        index = base.get('obj_num',0) - base.get('start_obj_num',0)
        index_key = 'obj_num'
    else:
        index = base.get(index_key,0)

    return index, index_key


def GetRNG(config, base, logger=None, tag=''):
    """Get the appropriate current rng according to whatever the current index_key is.

    If a logger is provided, then it will emit a warning if there is no current rng setup.

    Parameters:
        config:         The configuration dict for the current item being worked on.
        base:           The base configuration dict.
        logger:         If given, a logger object to log progress. [default: None]
        tag:            If given, an appropriate name for the current item to use in the
                        warning message. [default: '']

    Returns:
        either the appropriate rng for the current index_key or None
    """
    logger = LoggerWrapper(logger)
    if 'rng_index_key' in config:
        index_key = config['rng_index_key']
        if index_key not in valid_index_keys:
            raise GalSimConfigValueError("Invalid rng_index_key.", index_key, valid_index_keys)
    else:
        index, index_key = GetIndex(config, base)
    logger.debug("GetRNG for %s",index_key)

    rng_num = config.get('rng_num', 0)
    if rng_num != 0:
        if int(rng_num) != rng_num:
            raise GalSimConfigValueError("rng_num must be an integer", rng_num)
        rngs = base.get(index_key + '_rngs', None)
        if rngs is None:
            raise GalSimConfigError("rng_num is only allowed when image.random_seed is a list")
        if rng_num < 0 or rng_num > len(rngs):
            raise GalSimConfigError("rng_num is invalid.  Must be in [0,%d]"%(len(rngs)))
        rng = rngs[int(rng_num)]
    else:
        rng = base.get(index_key + '_rng', None)

    if rng is None:
        logger.debug("No index_key_rng.  Use base[rng]")
        rng = base.get('rng',None)

    if rng is None and logger:
        # Only report the warning the first time.
        rng_tag = tag + '_reported_no_rng'
        if rng_tag not in base:
            base[rng_tag] = True
            logger.warning("No base['rng'] available for %s.  Using /dev/urandom.",tag)

    return rng

def CleanConfig(config, keep_current=False):
    """Return a "clean" config dict without any leading-underscore values

    GalSim config dicts store a lot of ancillary information internally to help improve
    efficiency.  However, some of these are actually pointers to other places in the dict, so
    printing a config dict, or even what should be a small portion of one, can have infinite loops.

    This helper function is useful when debugging config processing to strip out all of these
    leading-underscore values, so that printing the dict is reasonable.

        >>> print(galsim.config.CleanConfig(config_dict))
    """
    if isinstance(config, dict):
        return { k : CleanConfig(config[k], keep_current) for k in config
                 if not (isinstance(k, str) and k[0] == '_')
                 and (keep_current or k != 'current') }
    elif isinstance(config, list):
        return [ CleanConfig(item, keep_current) for item in config ]
    else:
        return config

def SetDefaultExt(config, default_ext):
    """Set a default ext in a config 'file_name' field if appropriate.

    Parameters:
        config:         The configuration dict for the item that might need to be given
                        a default 'ext' value.
        default_ext:    The default extension to set in the config dict if one is not set.
    """
    if ( isinstance(config,dict) and 'type' in config and
        config['type'] == 'NumberedFile' and 'ext' not in config ):
        config['ext'] = default_ext


_sleep_mult = 1  # 1 second normally, but make it a variable, so I can change it when unit testing.
def RetryIO(func, args, ntries, file_name, logger):
    """A helper function to retry I/O commands
    """
    itry = 0
    while True:
        itry += 1
        try:
            ret = func(*args)
        except OSError as e:
            if itry == ntries:
                # Then this was the last try.  Just re-raise the exception.
                raise
            else:
                logger.warning('File %s: Caught OSError: %s',file_name,str(e))
                logger.warning('This is try %d/%d, so sleep for %d sec and try again.',
                               itry,ntries,itry)
                time.sleep(itry * _sleep_mult)
                continue
        else:
            break
    return ret

def get_cls_params(cls):
    """Get the _req_params, _opt_params, _single_params, and _takes_rng definitions
    from the given class (or function, but usually class).

    This handles the defaults of empty dict or list or False as appropriate.
    """
    req = getattr(cls,'_req_params', {})
    opt = getattr(cls,'_opt_params', {})
    single = getattr(cls,'_single_params', [])
    takes_rng = getattr(cls,'_takes_rng', False)

    return req, opt, single, takes_rng



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

from ..errors import GalSimDeprecationWarning

def depr(f, v, s1, s2=None):
    """A helper function for emitting a GalSimDeprecationWarning.

    Example usage:

    1. Normally, you would simply warn that a deprecated function has a new preferred syntax:

            depr('applyShear', 1.1, 'obj = obj.shear(..)')

    2. You can add extra information if you want to point out something about the new syntax:

            depr('draw', 1.1, "drawImage(..., method='no_pixel')",
                 'Note: drawImage has different args than draw did.  '
                 'Read the docs for the method keywords carefully.')

    3. If the deprecated function has no replacement, you can use '' for the first string.

            depr('calculateCovarianceMatrix', 1.3, '',
                 'This functionality has been removed. If you have a need for it, please open '
                 'an issue requesting the functionality.')
    """
    import warnings
    s = str(f)+' has been deprecated since GalSim version '+str(v)+'.'
    if s1:
        s += '  Use '+s1+' instead.'
    if s2:
        s += '  ' + s2
    warnings.warn(s, GalSimDeprecationWarning)

from . import correlatednoise
from . import randwalk
from . import integ



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import os

import galsim
from galsim_test_helpers import *

path, filename = os.path.split(__file__)
imgdir = os.path.join(path, "SBProfile_comparison_images") # Directory containing the reference
                                                           # images.

@timer
def test_exponential():
    """Test the generation of a specific exp profile against a known result.
    """
    re = 1.0
    # Note the factor below should really be 1.6783469900166605, but the value of 1.67839 is
    # retained here as it was used by SBParse to generate the original known result (this changed
    # in commit b77eb05ab42ecd31bc8ca03f1c0ae4ee0bc0a78b.
    # The value of this test for regression purposes is not harmed by retaining the old scaling, it
    # just means that the half light radius chosen for the test is not really 1, but 0.999974...
    r0 = re/1.67839
    savedImg = galsim.fits.read(os.path.join(imgdir, "exp_1.fits"))
    dx = 0.2
    myImg = galsim.ImageF(savedImg.bounds, scale=dx)
    myImg.setCenter(0,0)

    expon = galsim.Exponential(flux=1., scale_radius=r0)
    expon.drawImage(myImg,scale=dx, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Exponential disagrees with expected result")

    # Check with default_params
    expon = galsim.Exponential(flux=1., scale_radius=r0, gsparams=default_params)
    expon.drawImage(myImg,scale=dx, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Exponential with default_params disagrees with expected result")
    expon = galsim.Exponential(flux=1., scale_radius=r0, gsparams=galsim.GSParams())
    expon.drawImage(myImg,scale=dx, method="sb", use_true_center=False)
    np.testing.assert_array_almost_equal(
            myImg.array, savedImg.array, 5,
            err_msg="Using GSObject Exponential with GSParams() disagrees with expected result")

    # Use non-unity values.
    expon = galsim.Exponential(flux=1.7, scale_radius=0.91)
    gsp = galsim.GSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)
    expon2 = galsim.Exponential(flux=1.7, scale_radius=0.91, gsparams=gsp)
    assert expon2 != expon
    assert expon2 == expon.withGSParams(gsp)
    assert expon2 == expon.withGSParams(xvalue_accuracy=1.e-8, kvalue_accuracy=1.e-8)

    check_basic(expon, "Exponential")

    # Test photon shooting.
    do_shoot(expon,myImg,"Exponential")

    # Test kvalues
    do_kvalue(expon,myImg,"Exponential")

    # Check picklability
    check_pickle(expon, lambda x: x.drawImage(method='no_pixel'))
    check_pickle(expon)

    # Should raise an exception if both scale_radius and half_light_radius are provided.
    assert_raises(TypeError, galsim.Exponential, scale_radius=3, half_light_radius=1)
    # Or neither.
    assert_raises(TypeError, galsim.Exponential)


@timer
def test_exponential_properties():
    """Test some basic properties of the Exponential profile.
    """
    test_flux = 17.9
    test_scale = 1.8
    expon = galsim.Exponential(flux=test_flux, scale_radius=test_scale)
    # Check that we are centered on (0, 0)
    cen = galsim.PositionD(0, 0)
    np.testing.assert_equal(expon.centroid, cen)
    # Check Fourier properties
    np.testing.assert_almost_equal(expon.maxk, 10 / test_scale)
    np.testing.assert_almost_equal(expon.stepk, 0.37436747851 / test_scale)
    np.testing.assert_almost_equal(expon.kValue(cen), (1+0j) * test_flux)
    np.testing.assert_almost_equal(expon.flux, test_flux)
    import math
    np.testing.assert_almost_equal(expon.xValue(cen), 1./(2.*math.pi)*test_flux/test_scale**2)
    np.testing.assert_almost_equal(expon.xValue(cen), expon.max_sb)
    # Check input flux vs output flux
    for inFlux in np.logspace(-2, 2, 10):
        expon = galsim.Exponential(flux=inFlux, scale_radius=1.8)
        outFlux = expon.flux
        np.testing.assert_almost_equal(outFlux, inFlux)

    # Check that stepk and maxk scale correctly with radius
    expon2 = galsim.Exponential(flux=test_flux, scale_radius=5*test_scale)
    np.testing.assert_almost_equal(expon2.maxk, expon.maxk/5)
    np.testing.assert_almost_equal(expon2.stepk, expon.stepk/5)


@timer
def test_exponential_radii():
    """Test initialization of Exponential with different types of radius specification.
    """
    test_hlr = 1.8
    test_scale = 1.8

    import math
    # Test constructor using half-light-radius:
    test_gal = galsim.Exponential(flux=1., half_light_radius=test_hlr)
    hlr_sum = radial_integrate(test_gal, 0., test_hlr)
    print('hlr_sum = ',hlr_sum)
    np.testing.assert_almost_equal(
            hlr_sum, 0.5, decimal=4,
            err_msg="Error in Exponential constructor with half-light radius")

    # then test scale getter
    center = test_gal.xValue(galsim.PositionD(0,0))
    got_sr = test_gal.scale_radius
    ratio = test_gal.xValue(galsim.PositionD(got_sr,0)) / center
    print('scale ratio = ',ratio)
    np.testing.assert_almost_equal(
            ratio, np.exp(-1.0), decimal=4,
            err_msg="Error in scale_radius for Exponential constructed with half light radius")

    # Test constructor using scale radius:
    test_gal = galsim.Exponential(flux=1., scale_radius=test_scale)
    center = test_gal.xValue(galsim.PositionD(0,0))
    ratio = test_gal.xValue(galsim.PositionD(test_scale,0)) / center
    print('scale ratio = ',ratio)
    np.testing.assert_almost_equal(
            ratio, np.exp(-1.0), decimal=4,
            err_msg="Error in Exponential constructor with scale")

    # then test that image indeed has the correct HLR properties when radially integrated
    got_hlr = test_gal.half_light_radius
    hlr_sum = radial_integrate(test_gal, 0., got_hlr)
    print('hlr_sum (profile initialized with scale_radius) = ',hlr_sum)
    np.testing.assert_almost_equal(
            hlr_sum, 0.5, decimal=4,
            err_msg="Error in half light radius for Exponential initialized with scale_radius.")

    # Check that the getters don't work after modifying the original.
    test_gal_shear = test_gal.shear(g1=0.3, g2=0.1)
    assert_raises(AttributeError, getattr, test_gal_shear, "half_light_radius")
    assert_raises(AttributeError, getattr, test_gal_shear, "scale_radius")


@timer
def test_exponential_flux_scaling():
    """Test flux scaling for Exponential.
    """
    # decimal point to go to for parameter value comparisons
    param_decimal = 12
    test_flux = 17.9
    test_scale = 1.8

    # init with scale and flux only (should be ok given last tests)
    obj = galsim.Exponential(scale_radius=test_scale, flux=test_flux)
    obj *= 2.
    np.testing.assert_almost_equal(
        obj.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __imul__.")
    obj = galsim.Exponential(scale_radius=test_scale, flux=test_flux)
    obj /= 2.
    np.testing.assert_almost_equal(
        obj.flux, test_flux / 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __idiv__.")
    obj = galsim.Exponential(scale_radius=test_scale, flux=test_flux)
    obj2 = obj * 2.
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __rmul__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __rmul__ (result).")
    obj = galsim.Exponential(scale_radius=test_scale, flux=test_flux)
    obj2 = 2. * obj
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __mul__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux * 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __mul__ (result).")
    obj = galsim.Exponential(scale_radius=test_scale, flux=test_flux)
    obj2 = obj / 2.
    # First test that original obj is unharmed...
    np.testing.assert_almost_equal(
        obj.flux, test_flux, decimal=param_decimal,
        err_msg="Flux param inconsistent after __div__ (original).")
    # Then test new obj2 flux
    np.testing.assert_almost_equal(
        obj2.flux, test_flux / 2., decimal=param_decimal,
        err_msg="Flux param inconsistent after __div__ (result).")

@timer
def test_exponential_shoot():
    """Test Exponential with photon shooting.  Particularly the flux of the final image.
    """
    rng = galsim.BaseDeviate(1234)
    obj = galsim.Exponential(half_light_radius=3.5, flux=1.e4)
    im = galsim.Image(100,100, scale=1)
    im.setCenter(0,0)
    added_flux, photons = obj.drawPhot(im, poisson_flux=False, rng=rng.duplicate())
    print('obj.flux = ',obj.flux)
    print('added_flux = ',added_flux)
    print('photon fluxes = ',photons.flux.min(),'..',photons.flux.max())
    print('image flux = ',im.array.sum())
    assert np.isclose(added_flux, obj.flux)
    assert np.isclose(im.array.sum(), obj.flux)
    photons2 = obj.makePhot(poisson_flux=False, rng=rng)
    assert photons2 == photons, "Exponential makePhot not equivalent to drawPhot"


@timer
def test_ne():
    """Test base.py GSObjects for not-equals."""
    # Define some universal gsps
    gsp = galsim.GSParams(maxk_threshold=1.1e-3, folding_threshold=5.1e-3)

    # Exponential.  Params include half_light_radius, scale_radius, flux, gsparams
    # The following should all test unequal:
    gals = [galsim.Exponential(half_light_radius=1.0),
            galsim.Exponential(half_light_radius=1.1),
            galsim.Exponential(scale_radius=1.0),
            galsim.Exponential(half_light_radius=1.0, flux=1.1),
            galsim.Exponential(half_light_radius=1.0, gsparams=gsp)]
    check_all_diff(gals)


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np
import warnings

import galsim
from galsim_test_helpers import *


@timer
def test_nonlinearity_basic():
    """Check for overall sensible behavior of the nonlinearity routine."""
    # Make an image with non-trivially interesting scale and bounds.
    g = galsim.Gaussian(sigma=3.7)
    im = g.drawImage(scale=0.25)
    im.replaceNegative()  # For default float32 image, some values are -1.e-11, which messes
                          # up below tests that need I>=0.
    im.shift(dx=-5, dy=3)
    im_save = im.copy()

    # Basic - exceptions / bad usage (invalid function, does not return NumPy array).
    assert_raises(ValueError, im.applyNonlinearity, lambda x : 1.0)
    assert_raises(ValueError, im.applyNonlinearity, lambda x : np.array([1,2,3]))

    # Check for constant function as NLfunc
    im_new = im.copy()
    im_new.applyNonlinearity(lambda x: x**0.0)
    assert im_new.scale == im.scale
    assert im_new.wcs == im.wcs
    assert im_new.dtype == im.dtype
    assert im_new.bounds == im.bounds
    np.testing.assert_array_equal(
        im_new.array, np.ones_like(im),
        err_msg='Image not constant when the nonlinearity function is constant')

    # Check that calling a NLfunc with no parameter works
    NLfunc = lambda x: x + 0.001*(x**2)
    im_new = im.copy()
    im_new.applyNonlinearity(NLfunc)
    assert im_new.scale == im.scale
    assert im_new.wcs == im.wcs
    assert im_new.dtype == im.dtype
    assert im_new.bounds == im.bounds
    np.testing.assert_array_equal(
        im_new.array, im.array+0.001*((im.array)**2),
        err_msg = 'Nonlinearity function with no argument does not function as desired.')

    # Check that calling a NLfunc with a parameter works
    NLfunc = lambda x, beta: x + beta*(x**2)
    im_new = im.copy()
    im_new.applyNonlinearity(NLfunc, 0.001)
    assert im_new.scale == im.scale
    assert im_new.wcs == im.wcs
    assert im_new.dtype == im.dtype
    assert im_new.bounds == im.bounds
    np.testing.assert_array_equal(
        im_new.array, im.array + 0.001*((im.array)**2),
        err_msg = 'Nonlinearity function with one argument does not function as desired.')

    # Check that calling a NLfunc with multiple parameters works
    NLfunc = lambda x, beta1, beta2: x + beta1*(x**2) + beta2*(x**3)
    im_new = im.copy()
    im_new.applyNonlinearity(NLfunc, 0.001, -0.0001)
    assert im_new.scale == im.scale
    assert im_new.wcs == im.wcs
    assert im_new.dtype == im.dtype
    assert im_new.bounds == im.bounds
    np.testing.assert_array_equal(
        im_new.array, im.array + 0.001*((im.array)**2) -0.0001*((im.array)**3),
        err_msg = 'Nonlinearity function with multiple arguments does not function as desired')

    # Check for preservation for identity function as NLfunc.
    im_new = im.copy()
    im_new.applyNonlinearity(lambda x : x)
    assert im_new.scale == im.scale
    assert im_new.wcs == im.wcs
    assert im_new.dtype == im.dtype
    assert im_new.bounds == im.bounds
    np.testing.assert_array_equal(
        im_new.array, im.array,
        err_msg='Image not preserved when applying identity nonlinearity function')

    # Check that lambda func vs. LookupTable agree.
    max_val = np.max(im.array)
    x_vals = np.linspace(0.0, max_val, num=500)
    f_vals = x_vals + 0.1*(x_vals**2)
    lut = galsim.LookupTable(x=x_vals, f=f_vals)
    im1 = im.copy()
    im2 = im.copy()
    im1.applyNonlinearity(lambda x : x + 0.1*(x**2))
    im2.applyNonlinearity(lut)

    assert im1.scale == im.scale
    assert im1.wcs == im.wcs
    assert im1.dtype == im.dtype
    assert im1.bounds == im.bounds

    assert im2.scale == im.scale
    assert im2.wcs == im.wcs
    assert im2.dtype == im.dtype
    assert im2.bounds == im.bounds
    # Note, don't be quite as stringent as in previous test; there can be small interpolation
    # errors.
    np.testing.assert_array_almost_equal(
        im1.array, im2.array, 7,
        err_msg='Image differs when using LUT vs. lambda function')

    # Check that lambda func vs. interpolated function from SciPy agree
    # GalSim doesn't have SciPy dependence and this is NOT the preferred way to construct smooth
    # functions from tables but our routine can handle it anyway.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore",category=RuntimeWarning)
        from scipy import interpolate
    max_val = np.max(im.array)
    x_vals = np.linspace(0.0,max_val,num=500)
    y_vals = x_vals + 0.1*(x_vals**2)
    lut = interpolate.interp1d(x=x_vals,y=y_vals)
    im1 = im.copy()
    im2 = im.copy()
    im1.applyNonlinearity(lambda x: x + 0.1*(x**2))
    im2.applyNonlinearity(lut)

    assert im1.scale == im.scale
    assert im1.wcs == im.wcs
    assert im1.dtype == im.dtype
    assert im1.bounds == im.bounds

    assert im2.scale == im.scale
    assert im2.wcs == im.wcs
    assert im2.dtype == im.dtype
    assert im2.bounds == im.bounds

    # Note, don't be quite as stringent as in previous test; there can be small interpolation
    # errors.
    np.testing.assert_array_almost_equal(
        im1.array, im2.array, 7,
        err_msg="Image differs when using SciPy's interpolation vs. lambda function")


@timer
def test_recipfail_basic():
    """Check for overall sensible behavior of the reciprocity failure routine."""
    # Make an image with non-trivially interesting scale and bounds.
    g = galsim.Gaussian(sigma=3.7)
    im = g.drawImage(scale=0.25)
    im.replaceNegative(1.e-11)  # For default float32 image, some values are -1.e-11, which messes
                                # up below tests that need I>0.  They can't even handle I=0, so use
                                # a slightly positive value instead.
    im.shift(dx=-5, dy=3)
    im_save = im.copy()

    # Basic - exceptions / bad usage.
    assert_raises(ValueError, im.addReciprocityFailure, -1.0, 200, 1.0)
    assert_raises(ValueError, im.addReciprocityFailure, 1.0, -200, 1.0)
    assert_raises(ValueError, im.addReciprocityFailure, 1.0, 200, -1.0)

    # Preservation of data type / scale / bounds
    im_new = im.copy()
    im_new.addReciprocityFailure(exp_time=200., alpha=0.0065, base_flux=1.0)
    assert im_new.scale == im.scale
    assert im_new.wcs == im.wcs
    assert im_new.dtype == im.dtype
    assert im_new.bounds == im.bounds
    np.testing.assert_array_equal(
        im_save.array, im.array,
        err_msg = 'Input image was modified after addition of reciprocity failure')

    # Check for preservation for certain alpha.
    im_new = im.copy()
    im_new.addReciprocityFailure(exp_time=200., alpha=0.0, base_flux=1.0)
    assert im_new.scale == im.scale
    assert im_new.wcs == im.wcs
    assert im_new.dtype == im.dtype
    assert im_new.bounds == im.bounds
    np.testing.assert_array_equal(
        im_new.array, im.array,
        err_msg='Image not preserved when applying null reciprocity failure')

    # Check for proper scaling with alpha
    alpha1 = 0.006
    alpha2 = 0.007
    im1 = im.copy()
    im2 = im.copy()
    im1.addReciprocityFailure(exp_time=200.,alpha=alpha1, base_flux=1.0)
    im2.addReciprocityFailure(exp_time=200.,alpha=alpha2, base_flux=1.0)
    dim1 = im1.array/im.array
    dim2 = im2.array/im.array
    # We did new / old image, which should equal (old_image/normalization)^alpha.
    # The old image is the same, as is the factor inside the log.  So the log ratio log(dim2)/log(
    # dim1) should just be alpha2/alpha1
    expected_ratio = np.zeros(im.array.shape) + (alpha2/alpha1)

    assert im1.scale == im.scale
    assert im1.wcs == im.wcs
    assert im1.dtype == im.dtype
    assert im1.bounds == im.bounds

    assert im2.scale == im.scale
    assert im2.wcs == im.wcs
    assert im2.dtype == im.dtype
    assert im2.bounds == im.bounds

    np.testing.assert_array_almost_equal(
        np.log(dim2)/np.log(dim1), expected_ratio, 5,
        err_msg='Did not get expected change in reciprocity failure when varying alpha')

    # Check math is right
    alpha, exp_time, base_flux = 0.0065, 10.0, 5.0
    im_new = im.copy()
    im_new.addReciprocityFailure(alpha=alpha, exp_time=exp_time, base_flux=base_flux)
    assert im_new.scale == im.scale
    assert im_new.wcs == im.wcs
    assert im_new.dtype == im.dtype
    assert im_new.bounds == im.bounds
    np.testing.assert_array_almost_equal(
        (np.log10(im_new.array)-np.log10(im.array)), (alpha/np.log(10))*np.log10(im.array/ \
            (exp_time*base_flux)), 5,
        err_msg='Difference in images is not alpha times the log of original')

    # Check power law against logarithmic behavior
    alpha, exp_time, base_flux = 0.0065, 2.0, 4.0
    im_new = im.copy()
    im_new.addReciprocityFailure(alpha=alpha, exp_time=exp_time, base_flux=base_flux)
    assert im_new.scale == im.scale
    assert im_new.wcs == im.wcs
    assert im_new.dtype == im.dtype
    assert im_new.bounds == im.bounds
    np.testing.assert_array_almost_equal(
        im_new.array,im.array*(1+alpha*np.log10(im.array/(exp_time*base_flux))),6,
        err_msg='Difference between power law and log behavior')

    # If input image has negative values, then raise a warning.
    im_new.setValue(30, 30, -100)
    with assert_warns(galsim.GalSimWarning):
        im_new.addReciprocityFailure(alpha=alpha, exp_time=exp_time, base_flux=base_flux)


@timer
def test_quantize():
    """Check behavior of the image quantization routine."""
    # Choose a set of types.
    dtypes = [np.float64, np.float32]
    for dtype in dtypes:

        # Set up some array and image with this type.
        arr = np.arange(-10.2,9.8,0.5,dtype=dtype).reshape(5,8)
        image = galsim.Image(arr, scale=0.3, xmin=37, ymin=-14)
        image_q = image.copy()

        # Do quantization.
        image_q.quantize()

        # Check for correctness of numerical values.
        # Note that quantize uses np.round, which rounds x.5 to the nearest even number.
        # cf. http://docs.scipy.org/doc/numpy/reference/generated/numpy.around.html#numpy.around
        # This is different from floor(array+0.5), so don't use x.5 in the test array here.
        # For all other values, the two prescriptions should be equivalent.
        np.testing.assert_array_almost_equal(
            image_q.array, np.floor(image.array+0.5), decimal=8,
            err_msg='Array contents not as expected after quantization, for dtype=%s'%dtype)

        # Make sure quantizing an image with values that are already integers does nothing.
        save_image = image_q.copy()
        image_q.quantize()
        np.testing.assert_array_almost_equal(
            image_q.array, save_image.array, decimal=8,
            err_msg='Array contents of integer-valued array modified by quantization, '+
            'for dtype=%s'%dtype)

        # Check for preservation of WCS etc.
        assert image_q.scale == image.scale
        assert image_q.wcs == image.wcs
        assert image_q.dtype == image.dtype
        assert image_q.bounds == image.bounds


@timer
def test_IPC_basic():
    # Make an image with non-trivially interesting scale.
    g = galsim.Gaussian(sigma=3.7)
    im = g.drawImage(scale=0.25, dtype=float)
    im_save = im.copy()

    # Check for no IPC
    ipc_kernel = galsim.Image(3,3)
    ipc_kernel.setValue(2,2,1.0)
    im_new = im.copy()

    im_new.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='extend')
    np.testing.assert_array_equal(
        im_new.array, im.array,
        err_msg="Image is altered for no IPC with edge_treatment = 'extend'" )

    im_new.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='wrap')
    np.testing.assert_array_equal(
        im_new.array, im.array,
        err_msg="Image is altered for no IPC with edge_treatment = 'wrap'" )

    im_new.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='crop')
    np.testing.assert_array_equal(
        im_new.array, im.array,
        err_msg="Image is altered for no IPC with edge_treatment = 'crop'" )

    assert_raises(ValueError, im_new.applyIPC, galsim.Image(2,2,init_value=1))
    assert_raises(ValueError, im_new.applyIPC, galsim.Image(3,3,init_value=-1))
    assert_raises(ValueError, im_new.applyIPC, ipc_kernel * -1)
    assert_raises(ValueError, im_new.applyIPC, ipc_kernel, edge_treatment='invalid')

    # Test with a scalar fill_value
    fill_value = np.pi # a non-trivial one
    im_new.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='crop',fill_value=fill_value)

    #Input arrays and output arrays will differ at the edges for this option.
    np.testing.assert_array_equal(
        im_new.array[1:-1,1:-1], im.array[1:-1,1:-1],
        err_msg="Image is altered for no IPC with edge_treatment = 'crop' and with a fill_value" )
    # Check if the edges are filled with fill_value
    np.testing.assert_array_equal(
        im_new.array[0,:], fill_value,
        err_msg="Top edge is not filled with the correct value by applyIPC")
    np.testing.assert_array_equal(
        im_new.array[-1,:], fill_value,
        err_msg="Bottom edge is not filled with the correct value by applyIPC")
    np.testing.assert_array_equal(
        im_new.array[:,0], fill_value,
        err_msg="Left edge is not filled with the correct value by applyIPC")
    np.testing.assert_array_equal(
        im_new.array[:,-1], fill_value,
        err_msg="Left edge is not filled with the correct value by applyIPC")

    # Testing for flux conservation
    np.random.seed(1234)
    ipc_kernel = galsim.Image(abs(np.random.randn(3,3))) # a random kernel
    im_new = im.copy()
    # Set edges to zero since flux is not conserved at the edges otherwise
    im_new.array[0,:] = 0.0
    im_new.array[-1,:] = 0.0
    im_new.array[:,0] = 0.0
    im_new.array[:,-1] = 0.0
    with assert_warns(galsim.GalSimWarning):  # warn about the sum not being 1
        im_new.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='extend')
    np.testing.assert_almost_equal(im_new.array.sum(), im.array[1:-1,1:-1].sum(), 4,
        err_msg="Normalized IPC kernel does not conserve the total flux for 'extend' option.")

    # With kernel_normalization = False, it won't warn, but it also won't conserve flux.
    im_new = im.copy()
    im_new.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='extend', kernel_normalization=False)
    assert np.abs(im_new.array.sum() - im.array[1:-1,1:-1].sum()) > 1.e-8

    im_new = im.copy()
    ipc_kernel /= ipc_kernel.array.sum()  # Explicitly normalizing also avoids warning.
    im_new.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='wrap')
    np.testing.assert_almost_equal(im_new.array.sum(), im.array.sum(), 4,
        err_msg="Normalized IPC kernel does not conserve the total flux for 'wrap' option.")

    # Checking directionality
    ipc_kernel = galsim.Image(3,3)
    ipc_kernel.setValue(2,2,0.875)
    ipc_kernel.setValue(2,3,0.125)
    # This kernel should correspond to each pixel getting contribution from the pixel above it.
    im1 = im.copy()
    im1.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='crop')
    np.testing.assert_array_almost_equal(0.875*im.array[1:-1,1:-1]+0.125*im.array[2:,1:-1],
        im1.array[1:-1,1:-1], 7, err_msg="Difference in directionality for up kernel in applyIPC")
    # Checking for one pixel in the central bulk
    np.testing.assert_almost_equal(im1(2,2), 0.875*im(2,2)+0.125*im(2,3), 7,
        err_msg="Direction is not as intended for up kernel in applyIPC")

    ipc_kernel = galsim.Image(3,3)
    ipc_kernel.setValue(2,2,0.875)
    ipc_kernel.setValue(1,2,0.125)
    # This kernel should correspond to each pixel getting contribution from the pixel to its left.
    im1 = im.copy()
    im1.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='crop')
    np.testing.assert_array_almost_equal(im1.array[1:-1,1:-1], im1.array[1:-1,1:-1], 7,
        err_msg="Difference in directionality for left kernel in applyIPC")
    # Checking for one pixel in the central bulk
    np.testing.assert_almost_equal(im1(2,3), 0.875*im(2,3)+0.125*im(2,2), 7,
        err_msg="Direction is not as intended for left kernel in applyIPC")

    # Check using GalSim's native Convolve routine for GSObjects for a realisitic kernel
    ipc_kernel = galsim.Image(np.array(
        [[0.01,0.1,0.01],
         [0.1,1.0,0.1],
         [0.01,0.1,0.01]]))
    ipc_kernel /= ipc_kernel.array.sum()
    ipc_kernel_int = galsim.InterpolatedImage(ipc_kernel,x_interpolant='nearest',scale=im.scale)
    im1 = im.copy()
    im1.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='crop',kernel_normalization=False)
    im2 = im.copy()
    im2_int = galsim.InterpolatedImage(im2,x_interpolant='nearest')
    ipc_kernel_int = galsim.InterpolatedImage(ipc_kernel,x_interpolant='nearest',scale=im.scale)
    im_int = galsim.Convolve(ipc_kernel_int,im2_int,real_space=False)
    im_int.drawImage(im2,method='no_pixel',scale=im.scale)
    np.testing.assert_array_almost_equal(im1.array,im2.array,6,
        err_msg="Output of applyIPC does not match the output from Convolve")

    # SciPy is going to emit a warning that we don't want to worry about, so let's deliberately
    # ignore it by going into a `catch_warnings` context.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore",category=RuntimeWarning)
        warnings.simplefilter("ignore")
        from scipy import signal

        # Generate an arbitrary kernel
        np.random.seed(2345)
        ipc_kernel = galsim.Image(abs(np.random.randn(3,3)))
        ipc_kernel /= ipc_kernel.array.sum()
        # Convolution requires the kernel to be flipped up-down and left-right.
        im_new = im.copy()
        im_new.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='extend',
                        kernel_normalization=False)
        np.testing.assert_array_almost_equal(
            im_new.array, signal.convolve2d(im.array, np.flipud(np.fliplr(ipc_kernel.array)),
                                            mode='same', boundary='fill'), 7,
            err_msg="Image differs from SciPy's result using `mode='same'` and "
                "`boundary='fill`.")

        im_new = im.copy()
        im_new.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='crop',
                        kernel_normalization=False)
        np.testing.assert_array_almost_equal(
            im_new.array[1:-1,1:-1], signal.convolve2d(im.array,
                                                        np.fliplr(np.flipud(ipc_kernel.array)),
                                                        mode='valid', boundary = 'fill'), 7,
            err_msg="Image differs from SciPy's result using `mode=valid'` and "
                "`boundary='fill'`.")

        im_new = im.copy()
        im_new.applyIPC(IPC_kernel=ipc_kernel, edge_treatment='wrap',
                        kernel_normalization=False)
        np.testing.assert_array_almost_equal(
            im_new.array, signal.convolve2d(im.array, np.fliplr(np.flipud(ipc_kernel.array)),
                                            mode='same', boundary='wrap'), 7,
            err_msg="Image differs from SciPy's result using `mode=same'` and "
                "boundary='wrap'`.")


@timer
def test_Persistence_basic():
    # Make an image with non-trivially interesting scale and bounds.
    g = galsim.Gaussian(sigma=3.7,flux=1000.)
    im = g.drawImage(scale=0.25)
    im.shift(dx=-5, dy=3)
    im_save = im.copy()

    # Make 3 more images to act as previous images
    dx = [0.0, 1.0, -4.0]
    dy = [-1.0, 0.0, 2.0]
    im_prev = []
    for i in range(3):
        g = galsim.Gaussian(sigma=3.7,flux=1000.)
        im_prev += [g.drawImage(scale=0.25)]
        im_prev[i].shift(dx=dx[i],dy=dy[i])

    # Test for zero coefficient
    im1 = im.copy()
    im1.applyPersistence(imgs=im_prev,coeffs=[0.0]*len(im_prev))
    np.testing.assert_array_equal(im1.array, im.array,
        err_msg="Images do not agree when the persistence coefficients are all zeros.")

    # Test for a constant array
    im1 = im.copy()
    im2 = im.copy()
    for img in im_prev:
      im1 += 0.1*img
    im2.applyPersistence(imgs=im_prev, coeffs=0.1*np.ones_like(im_prev))
    np.testing.assert_array_equal(im1.array, im2.array,
        err_msg="Images differ when the persistence coefficients is a constant array.")

    # Test for identical copies of same image
    im_new = im.copy()
    n_im = 3
    im_new.applyPersistence(imgs=[im]*n_im, coeffs=0.5**np.linspace(1,n_im,n_im)) #0.5,0.25,0.125
    np.testing.assert_array_almost_equal(im_new.array, im.array*(2.-1./2**n_im),7, #sum of GP terms
            err_msg="Images differ when identical copies of the same image persist.")

    # Test for different lengths of imgs and coeffs
    im_new = im.copy()
    with assert_raises(ValueError):
        im_new.applyPersistence(im_prev, [0.2, 0.3])

    # Test for a single image and coeffs as a float
    im_new = im.copy()
    with assert_raises(TypeError):
        im_new.applyPersistence(im_prev[0], 1.0)

    # Testing the multiple images and varying coeffs
    im1 = im.copy()
    im2 = im.copy()
    im1.applyPersistence(imgs=im_prev, coeffs=np.linspace(1,len(im_prev), len(im_prev)))
    for i in range(len(im_prev)):
        im2 += (i+1)*im_prev[i]
    np.testing.assert_array_equal(im1.array, im2.array,
        err_msg="'applyPersistence' routine fails for multiple images with varying coefficients.")


if __name__ == "__main__":
    runtests(__file__)
